@@ -5,25 +5,25 @@ import atexit
 import json
 import os
 import re
+import sys
 from contextlib import suppress
 from functools import partial
 from gettext import gettext as _
+from time import monotonic
 from typing import (
-    Any, Callable, Container, Dict, Iterable, Iterator, List, Optional, Tuple,
-    Union
+    Any, Callable, Container, Dict, Iterable, Iterator, List, Optional, Set,
+    Tuple, Union
 )
 from weakref import WeakValueDictionary
 
-from .child import (
-    cached_process_data, cwd_of_process, default_env, set_default_env
-)
+from .child import cached_process_data, default_env, set_default_env
 from .cli import create_opts, parse_args
 from .cli_stub import CLIOptions
 from .conf.utils import BadLine, KeyAction, to_cmdline
 from .config import common_opts_as_dict, prepare_config_file_for_editing
 from .constants import (
-    appname, config_dir, is_macos, is_wayland, kitty_exe, logo_png_file,
-    supports_primary_selection, website_url
+    appname, cache_dir, config_dir, is_macos, is_wayland, kitty_exe,
+    logo_png_file, supports_primary_selection, website_url
 )
 from .fast_data_types import (
     CLOSE_BEING_CONFIRMED, GLFW_MOD_ALT, GLFW_MOD_CONTROL, GLFW_MOD_SHIFT,
@@ -56,12 +56,13 @@ from .tabs import (
 from .types import _T, AsyncResponse, SingleKey, WindowSystemMouseEvent, ac
 from .typing import PopenType, TypedDict
 from .utils import (
-    func_name, get_editor, get_new_os_window_size, get_primary_selection,
-    is_path_in_temp_dir, log_error, open_url, parse_address_spec,
-    parse_uri_list, platform_window_id, remove_socket_file, safe_print,
-    set_primary_selection, single_instance, startup_notification_handler, which
+    cleanup_ssh_control_masters, func_name, get_editor, get_new_os_window_size,
+    get_primary_selection, is_path_in_temp_dir, less_version, log_error,
+    macos_version, open_url, parse_address_spec, parse_uri_list,
+    platform_window_id, remove_socket_file, safe_print, set_primary_selection,
+    single_instance, startup_notification_handler, which
 )
-from .window import CommandOutput, MatchPatternType, Window
+from .window import CommandOutput, CwdRequest, Window
 
 
 class OSWindowDict(TypedDict):
@@ -229,7 +230,7 @@ class Boss:
         global_shortcuts: Dict[str, SingleKey]
     ):
         set_layout_options(opts)
-        self.cocoa_application_launched = False
+        self.update_check_started = False
         self.clipboard_buffers: Dict[str, str] = {}
         self.update_check_process: Optional['PopenType[bytes]'] = None
         self.window_id_map: WeakValueDictionary[int, Window] = WeakValueDictionary()
@@ -273,9 +274,9 @@ class Boss:
         for sc in self.global_shortcuts.values():
             self.keymap.pop(sc, None)
 
-    def startup_first_child(self, os_window_id: Optional[int]) -> None:
-        startup_sessions = create_sessions(get_options(), self.args, default_session=get_options().startup_session)
-        for startup_session in startup_sessions:
+    def startup_first_child(self, os_window_id: Optional[int], startup_sessions: Iterable[Session] = ()) -> None:
+        si = startup_sessions or create_sessions(get_options(), self.args, default_session=get_options().startup_session)
+        for startup_session in si:
             self.add_os_window(startup_session, os_window_id=os_window_id)
             os_window_id = None
             if self.args.start_as != 'normal':
@@ -340,42 +341,19 @@ class Boss:
             yield from tab
 
     def match_windows(self, match: str) -> Iterator[Window]:
-        try:
-            field, exp = match.split(':', 1)
-        except ValueError:
+        if match == 'all':
+            yield from self.all_windows
             return
-        if field == 'num':
-            tab = self.active_tab
-            if tab is not None:
-                try:
-                    w = tab.get_nth_window(int(exp))
-                except Exception:
-                    return
-                if w is not None:
-                    yield w
-            return
-        if field == 'recent':
-            tab = self.active_tab
-            if tab is not None:
-                try:
-                    num = int(exp)
-                except Exception:
-                    return
-                w = self.window_id_map.get(tab.nth_active_window_id(num))
-                if w is not None:
-                    yield w
-            return
-        if field != 'env':
-            pat: MatchPatternType = re.compile(exp)
-        else:
-            kp, vp = exp.partition('=')[::2]
-            if vp:
-                pat = re.compile(kp), re.compile(vp)
-            else:
-                pat = re.compile(kp), None
-        for window in self.all_windows:
-            if window.matches(field, pat):
-                yield window
+        from .search_query_parser import search
+        tab = self.active_tab
+
+        def get_matches(location: str, query: str, candidates: Set[int]) -> Set[int]:
+            return {wid for wid in candidates if self.window_id_map[wid].matches_query(location, query, tab)}
+
+        for wid in search(match, (
+                'id', 'title', 'pid', 'cwd', 'cmdline', 'num', 'env', 'recent', 'state'
+        ), set(self.window_id_map), get_matches):
+            yield self.window_id_map[wid]
 
     def tab_for_window(self, window: Window) -> Optional[Tab]:
         for tab in self.all_tabs:
@@ -385,61 +363,42 @@ class Boss:
         return None
 
     def match_tabs(self, match: str) -> Iterator[Tab]:
-        try:
-            field, exp = match.split(':', 1)
-        except ValueError:
-            return
-        pat = re.compile(exp)
+        if match == 'all':
+            return self.all_tabs
+        from .search_query_parser import search
+        tm = self.active_tab_manager
+        tim = {t.id: t for t in self.all_tabs}
+
+        def get_matches(location: str, query: str, candidates: Set[int]) -> Set[int]:
+            return {wid for wid in candidates if tim[wid].matches_query(location, query, tm)}
+
         found = False
-        if field in ('title', 'id'):
-            for tab in self.all_tabs:
-                if tab.matches(field, pat):
-                    yield tab
-                    found = True
-        elif field in ('window_id', 'window_title'):
-            wf = field.split('_')[1]
-            tabs = {self.tab_for_window(w) for w in self.match_windows(f'{wf}:{exp}')}
-            for q in tabs:
-                if q:
-                    found = True
-                    yield q
-        elif field == 'index':
-            tm = self.active_tab_manager
-            if tm is not None and len(tm.tabs) > 0:
-                idx = (int(pat.pattern) + len(tm.tabs)) % len(tm.tabs)
-                found = True
-                yield tm.tabs[idx]
-        elif field == 'recent':
-            tm = self.active_tab_manager
-            if tm is not None and len(tm.tabs) > 0:
-                try:
-                    num = int(exp)
-                except Exception:
-                    return
-                q = tm.nth_active_tab(num)
-                if q is not None:
-                    found = True
-                    yield q
+        for tid in search(match, (
+                'id', 'index', 'title', 'window_id', 'window_title', 'pid', 'cwd', 'env', 'cmdline', 'recent', 'state'
+        ), set(tim), get_matches):
+            found = True
+            yield tim[tid]
+
         if not found:
             tabs = {self.tab_for_window(w) for w in self.match_windows(match)}
             for q in tabs:
                 if q:
                     yield q
 
-    def set_active_window(self, window: Window, switch_os_window_if_needed: bool = False) -> Optional[int]:
+    def set_active_window(self, window: Window, switch_os_window_if_needed: bool = False, for_keep_focus: bool = False) -> Optional[int]:
         for os_window_id, tm in self.os_window_map.items():
             for tab in tm:
                 for w in tab:
                     if w.id == window.id:
                         if tab is not self.active_tab:
-                            tm.set_active_tab(tab)
-                        tab.set_active_window(w)
+                            tm.set_active_tab(tab, for_keep_focus=window.tabref() if for_keep_focus else None)
+                        tab.set_active_window(w, for_keep_focus=window if for_keep_focus else None)
                         if switch_os_window_if_needed and current_os_window() != os_window_id:
                             focus_os_window(os_window_id, True)
                         return os_window_id
         return None
 
-    def _new_os_window(self, args: Union[SpecialWindowInstance, Iterable[str]], cwd_from: Optional[int] = None) -> int:
+    def _new_os_window(self, args: Union[SpecialWindowInstance, Iterable[str]], cwd_from: Optional[CwdRequest] = None) -> int:
         if isinstance(args, SpecialWindowInstance):
             sw: Optional[SpecialWindowInstance] = args
         else:
@@ -461,12 +420,14 @@ class Boss:
     @ac('win', 'New OS Window with the same working directory as the currently active window')
     def new_os_window_with_cwd(self, *args: str) -> None:
         w = self.active_window_for_cwd
-        cwd_from = w.child.pid_for_cwd if w is not None else None
-        self._new_os_window(args, cwd_from)
+        self._new_os_window(args, CwdRequest(w))
 
-    def new_os_window_with_wd(self, wd: str) -> None:
-        special_window = SpecialWindow(None, cwd=wd)
-        self._new_os_window(special_window)
+    def new_os_window_with_wd(self, wd: Union[str, List[str]], str_is_multiple_paths: bool = False) -> None:
+        if isinstance(wd, str):
+            wd = wd.split(os.pathsep) if str_is_multiple_paths else [wd]
+        for path in wd:
+            special_window = SpecialWindow(None, cwd=path)
+            self._new_os_window(special_window)
 
     def add_child(self, window: Window) -> None:
         assert window.child.pid is not None and window.child.child_fd is not None
@@ -546,6 +507,10 @@ class Boss:
             from .cli_stub import CLIOptions
             startup_id = data.get('startup_id')
             args, rest = parse_args(data['args'][1:], result_class=CLIOptions)
+            cmdline_args_for_open = data.get('cmdline_args_for_open')
+            if cmdline_args_for_open:
+                self.launch_urls(*cmdline_args_for_open, no_replace_window=True)
+                return None
             args.args = rest
             opts = create_opts(args)
             if args.session == '-':
@@ -667,6 +632,15 @@ class Boss:
         if tab:
             self.confirm_tab_close(tab)
 
+    @ac('tab', 'Close all the tabs in the current OS window other than the currently active tab')
+    def close_other_tabs_in_os_window(self) -> None:
+        tm = self.active_tab_manager
+        if tm is not None and len(tm.tabs) > 1:
+            active_tab = self.active_tab
+            for tab in tm:
+                if tab is not active_tab:
+                    self.close_tab(tab)
+
     def confirm(
         self, msg: str,  # can contain newlines and ANSI formatting
         callback: Callable[..., None],  # called with True or False and *args
@@ -677,8 +651,9 @@ class Boss:
     ) -> None:
         def callback_(res: Dict[str, Any], x: int, boss: Boss) -> None:
             callback(res.get('response') == 'y', *args)
-        self._run_kitten('ask', ['--type=yesno', '--message', msg, '--default', 'y' if confirm_on_accept else 'n'],
-                         window=window, custom_callback=callback_, default_data={'response': 'y' if confirm_on_cancel else 'n'})
+        self.run_kitten_with_metadata(
+            'ask', ['--type=yesno', '--message', msg, '--default', 'y' if confirm_on_accept else 'n'],
+            window=window, custom_callback=callback_, default_data={'response': 'y' if confirm_on_cancel else 'n'})
 
     def choose(
         self, msg: str,  # can contain newlines and ANSI formatting
@@ -694,7 +669,19 @@ class Boss:
             cmd += ['-d', default]
         for c in choices:
             cmd += ['-c', c]
-        self._run_kitten('ask', cmd, window=window, custom_callback=callback_, default_data={'response': ''})
+        self.run_kitten_with_metadata('ask', cmd, window=window, custom_callback=callback_, default_data={'response': ''})
+
+    def get_line(
+        self, msg: str,  # can contain newlines and ANSI formatting
+        callback: Callable[..., None],  # called with the answer or empty string when aborted
+        window: Optional[Window] = None,  # the window associated with the confirmation
+        prompt: str = '> ',
+        is_password: bool = False
+    ) -> None:
+        def callback_(res: Dict[str, Any], x: int, boss: Boss) -> None:
+            callback(res.get('response') or '')
+        cmd = ['--type', 'password' if is_password else 'line', '--message', msg, '--prompt', prompt]
+        self.run_kitten_with_metadata('ask', cmd, window=window, custom_callback=callback_, default_data={'response': ''})
 
     def confirm_tab_close(self, tab: Tab) -> None:
         x = get_options().confirm_os_window_close
@@ -742,13 +729,20 @@ class Boss:
     def toggle_macos_secure_keyboard_entry(self) -> None:
         toggle_secure_input()
 
-    def start(self, first_os_window_id: int) -> None:
+    def start(self, first_os_window_id: int, startup_sessions: Iterable[Session]) -> None:
         if not getattr(self, 'io_thread_started', False):
             self.child_monitor.start()
             self.io_thread_started = True
-            self.startup_first_child(first_os_window_id)
+            urls: List[str] = getattr(sys, 'cmdline_args_for_open', [])
+            if urls:
+                delattr(sys, 'cmdline_args_for_open')
+                sess = create_sessions(get_options(), self.args, special_window=SpecialWindow([kitty_exe(), '+runpy', 'input()']))
+                self.startup_first_child(first_os_window_id, startup_sessions=tuple(sess))
+                self.launch_urls(*urls)
+            else:
+                self.startup_first_child(first_os_window_id, startup_sessions=startup_sessions)
 
-        if get_options().update_check_interval > 0 and not hasattr(self, 'update_check_started'):
+        if get_options().update_check_interval > 0 and not self.update_check_started and getattr(sys, 'frozen', False):
             from .update_check import run_update_check
             run_update_check(get_options().update_check_interval * 60 * 60)
             self.update_check_started = True
@@ -772,14 +766,15 @@ class Boss:
         See :sc:`reset_terminal <reset_terminal>` for details. For example::
 
             # Reset the terminal
-            map kitty_mod+f9 clear_terminal reset active
+            map f1 clear_terminal reset active
             # Clear the terminal screen by erasing all contents
-            map kitty_mod+f10 clear_terminal clear active
+            map f1 clear_terminal clear active
             # Clear the terminal scrollback by erasing it
-            map kitty_mod+f11 clear_terminal scrollback active
+            map f1 clear_terminal scrollback active
             # Scroll the contents of the screen into the scrollback
-            map kitty_mod+f12 clear_terminal scroll active
-
+            map f1 clear_terminal scroll active
+            # Clear everything up to the line with the cursor
+            map f1 clear_terminal to_cursor active
         ''')
     def clear_terminal(self, action: str, only_active: bool) -> None:
         if only_active:
@@ -789,17 +784,21 @@ class Boss:
                 windows.append(w)
         else:
             windows = list(self.all_windows)
-        reset = action == 'reset'
-        how = 3 if action == 'scrollback' else 2
-        for w in windows:
-            if action == 'scroll':
-                w.screen.scroll_until_cursor()
-                continue
-            w.screen.cursor.x = w.screen.cursor.y = 0
-            if reset:
-                w.screen.reset()
-            else:
-                w.screen.erase_in_display(how, False)
+        if action == 'reset':
+            for w in windows:
+                w.clear_screen(reset=True, scrollback=True)
+        elif action == 'scrollback':
+            for w in windows:
+                w.clear_screen(scrollback=True)
+        elif action == 'clear':
+            for w in windows:
+                w.clear_screen()
+        elif action == 'scroll':
+            for w in windows:
+                w.scroll_prompt_to_top()
+        elif action == 'to_cursor':
+            for w in windows:
+                w.scroll_prompt_to_top(clear_scrollback=True)
 
     def increase_font_size(self) -> None:  # legacy
         cfs = global_font_size()
@@ -1048,10 +1047,11 @@ class Boss:
             self.mouse_handler(ev)
 
     def select_window_in_tab_using_overlay(self, tab: Tab, msg: str, only_window_ids: Container[int] = ()) -> Optional[Window]:
-        windows = tuple((w.id, w.title) for i, w in tab.windows.iter_windows_with_number(only_visible=False)
-                        if not only_window_ids or w.id in only_window_ids)
+        windows = tuple((None, f'Current window: {w.title}' if w is self.active_window else w.title)
+                        if only_window_ids and w.id not in only_window_ids else (w.id, w.title)
+                        for i, w in tab.windows.iter_windows_with_number(only_visible=False))
         if len(windows) < 1:
-            self.visual_window_select_action_trigger(windows[0][0] if windows else 0)
+            self.visual_window_select_action_trigger(windows[0][0] if windows and windows[0][0] is not None else 0)
             if get_options().enable_audio_bell:
                 ring_bell()
             return None
@@ -1062,7 +1062,8 @@ class Boss:
             self.current_visual_select = None
             if cvs and q is cvs:
                 q.trigger(ans if isinstance(ans, int) else 0)
-        return self.choose_entry(msg, windows, chosen)
+
+        return self.choose_entry(msg, windows, chosen, hints_args=('--hints-offset=0', '--alphabet', get_options().visual_window_select_characters.lower()))
 
     @ac('win', '''
         Resize the active window interactively
@@ -1073,7 +1074,7 @@ class Boss:
         w = self.active_window
         if w is None:
             return
-        overlay_window = self._run_kitten('resize_window', args=[
+        overlay_window = self.run_kitten_with_metadata('resize_window', args=[
             f'--horizontal-increment={get_options().window_resize_step_cells}',
             f'--vertical-increment={get_options().window_resize_step_lines}'
         ])
@@ -1174,6 +1175,8 @@ class Boss:
                         if len(actions) > 1:
                             self.drain_actions(list(actions[1:]), window_for_dispatch, dispatch_type)
                 except Exception as e:
+                    import traceback
+                    traceback.print_exc()
                     self.show_error('Key action failed', f'{actions[0].pretty()}\n{e}')
                     consumed = True
         return consumed
@@ -1206,8 +1209,13 @@ class Boss:
             if w is not None:
                 text = data.decode('utf-8', 'replace')
                 if mime == 'text/uri-list':
-                    text = '\n'.join(parse_uri_list(text))
-                w.paste(text)
+                    urls = parse_uri_list(text)
+                    if w.at_prompt:
+                        import shlex
+                        text = ' '.join(map(shlex.quote, urls))
+                    else:
+                        text = '\n'.join(urls)
+                w.paste_text(text)
 
     @ac('win', 'Focus the nth OS window')
     def nth_os_window(self, num: int = 1) -> None:
@@ -1293,6 +1301,7 @@ class Boss:
             s.close()
 
     def display_scrollback(self, window: Window, data: Union[bytes, str], input_line_number: int = 0, title: str = '', report_cursor: bool = True) -> None:
+
         def prepare_arg(x: str) -> str:
             x = x.replace('INPUT_LINE_NUMBER', str(input_line_number))
             x = x.replace('CURSOR_LINE', str(window.screen.cursor.y + 1) if report_cursor else '0')
@@ -1308,6 +1317,15 @@ class Boss:
         tab = self.active_tab
         if tab is not None:
             bdata = data.encode('utf-8') if isinstance(data, str) else data
+            if is_macos and cmd[0] == '/usr/bin/less' and macos_version()[:2] < (12, 3):
+                # the system less before macOS 12.3 barfs up OSC codes, so sanitize them ourselves
+                sentinel = os.path.join(cache_dir(), 'less-is-new-enough')
+                if not os.path.exists(sentinel):
+                    if less_version(cmd[0]) >= 581:
+                        open(sentinel, 'w').close()
+                    else:
+                        bdata = re.sub(br'\x1b\].*?\x1b\\', b'', bdata)
+
             tab.new_special_window(
                 SpecialWindow(cmd, bdata, title or _('History'), overlay_for=window.id, cwd=window.cwd_of_child),
                 copy_colors_from=self.active_window
@@ -1321,16 +1339,7 @@ class Boss:
         cmd = [kitty_exe(), '+runpy', 'import os, sys, time; time.sleep(0.05); os.execvp(sys.argv[1], sys.argv[1:])'] + get_editor(get_options()) + [confpath]
         self.new_os_window(*cmd)
 
-    def get_output(self, source_window: Window, num_lines: Optional[int] = 1) -> str:
-        output = ''
-        s = source_window.screen
-        if num_lines is None:
-            num_lines = s.lines
-        for i in range(min(num_lines, s.lines)):
-            output += str(s.linebuf.line(i))
-        return output
-
-    def _run_kitten(
+    def run_kitten_with_metadata(
         self,
         kitten: str,
         args: Iterable[str] = (),
@@ -1393,7 +1402,8 @@ class Boss:
                         'OVERLAID_WINDOW_COLS': str(w.screen.columns),
                     },
                     cwd=w.cwd_of_child,
-                    overlay_for=w.id
+                    overlay_for=w.id,
+                    overlay_behind=end_kitten.has_ready_notification,
                 ),
                 copy_colors_from=w
             )
@@ -1406,22 +1416,21 @@ class Boss:
                         action_on_removal(wid, self)
                 overlay_window.actions_on_removal.append(callback_wrapper)
             return overlay_window
+    _run_kitten = run_kitten_with_metadata
 
     @ac('misc', 'Run the specified kitten. See :doc:`/kittens/custom` for details')
     def kitten(self, kitten: str, *kargs: str) -> None:
-        self._run_kitten(kitten, kargs)
+        self.run_kitten_with_metadata(kitten, kargs)
 
     def run_kitten(self, kitten: str, *args: str) -> None:
-        self._run_kitten(kitten, args)
+        self.run_kitten_with_metadata(kitten, args)
 
     def on_kitten_finish(
         self, target_window_id: int, end_kitten: Callable[[Dict[str, Any], int, 'Boss'], None],
         source_window: Window,
         default_data: Optional[Dict[str, Any]] = None
     ) -> None:
-        output = self.get_output(source_window, num_lines=None)
-        from kittens.runner import deserialize
-        data = deserialize(output)
+        data, source_window.kitten_result = source_window.kitten_result, None
         if data is None:
             data = default_data
         if data is not None:
@@ -1429,7 +1438,7 @@ class Boss:
 
     @ac('misc', 'Input an arbitrary unicode character. See :doc:`/kittens/unicode_input` for details.')
     def input_unicode_character(self) -> None:
-        self._run_kitten('unicode_input')
+        self.run_kitten_with_metadata('unicode_input')
 
     @ac('tab', 'Change the title of the active tab')
     def set_tab_title(self) -> None:
@@ -1438,7 +1447,7 @@ class Boss:
             args = [
                 '--name=tab-title', '--message', _('Enter the new title for this tab below.'),
                 '--default', tab.name or tab.title, 'do_set_tab_title', str(tab.id)]
-            self._run_kitten('ask', args)
+            self.run_kitten_with_metadata('ask', args)
 
     def do_set_tab_title(self, title: str, tab_id: int) -> None:
         tm = self.active_tab_manager
@@ -1450,7 +1459,12 @@ class Boss:
                     break
 
     def show_error(self, title: str, msg: str) -> None:
-        self._run_kitten('show_error', args=['--title', title], input_data=msg)
+        ec = sys.exc_info()
+        tb = ''
+        if ec != (None, None, None):
+            import traceback
+            tb = traceback.format_exc()
+        self.run_kitten_with_metadata('show_error', args=['--title', title], input_data=json.dumps({'msg': msg, 'tb': tb}))
 
     @ac('mk', 'Create a new marker')
     def create_marker(self) -> None:
@@ -1470,7 +1484,7 @@ class Boss:
                     except Exception as err:
                         self.show_error(_('Invalid marker specification'), str(err))
 
-            self._run_kitten('ask', [
+            self.run_kitten_with_metadata('ask', [
                 '--name=create-marker', '--message',
                 _('Create marker, for example:\ntext 1 ERROR\nSee {}\n').format(website_url('marks'))
                 ],
@@ -1513,6 +1527,10 @@ class Boss:
         if tab:
             tab.set_active_window(window_id)
 
+    def open_kitty_website(self) -> None:
+        self.open_url(website_url())
+
+    @ac('misc', 'Open the specified URL')
     def open_url(self, url: str, program: Optional[Union[str, List[str]]] = None, cwd: Optional[str] = None) -> None:
         if not url:
             return
@@ -1532,7 +1550,7 @@ class Boss:
 
     @ac('misc', 'Click a URL using the keyboard')
     def open_url_with_hints(self) -> None:
-        self._run_kitten('hints')
+        self.run_kitten_with_metadata('hints')
 
     def drain_actions(self, actions: List[KeyAction], window_for_dispatch: Optional[Window] = None, dispatch_type: str = 'KeyPress') -> None:
 
@@ -1557,7 +1575,7 @@ class Boss:
         if text:
             w = self.active_window
             if w is not None:
-                w.paste(text)
+                w.paste_with_actions(text)
 
     @ac('cp', 'Paste from the clipboard to the active window')
     def paste_from_clipboard(self) -> None:
@@ -1570,7 +1588,7 @@ class Boss:
     def current_primary_selection_or_clipboard(self) -> str:
         return get_primary_selection() if supports_primary_selection else get_clipboard_string()
 
-    @ac('cp', 'Paste from the clipboard to the active window')
+    @ac('cp', 'Paste from the primary selection, if present, otherwise the clipboard to the active window')
     def paste_from_selection(self) -> None:
         text = self.current_primary_selection_or_clipboard()
         self.paste_to_active_window(text)
@@ -1682,7 +1700,7 @@ class Boss:
         self, cmd: List[str],
         window: Optional[Window] = None,
         stdin: Optional[str] = None,
-        cwd_from: Optional[int] = None,
+        cwd_from: Optional[CwdRequest] = None,
         as_overlay: bool = False
     ) -> SpecialWindowInstance:
         w = window or self.active_window
@@ -1704,7 +1722,7 @@ class Boss:
         cwd: Optional[str] = None,
         env: Optional[Dict[str, str]] = None,
         stdin: Optional[bytes] = None,
-        cwd_from: Optional[int] = None
+        cwd_from: Optional[CwdRequest] = None
     ) -> None:
         import subprocess
         env = env or None
@@ -1714,7 +1732,7 @@ class Boss:
             env = env_
         if cwd_from:
             with suppress(Exception):
-                cwd = cwd_of_process(cwd_from)
+                cwd = cwd_from.cwd_of_child
 
         if stdin:
             r, w = safe_pipe(False)
@@ -1732,7 +1750,7 @@ class Boss:
     def pipe(self, source: str, dest: str, exe: str, *args: str) -> Optional[Window]:
         cmd = [exe] + list(args)
         window = self.active_window
-        cwd_from = window.child.pid_for_cwd if window else None
+        cwd_from = CwdRequest(window) if window else None
 
         def create_window() -> SpecialWindowInstance:
             return self.special_window_for_cmd(
@@ -1760,7 +1778,7 @@ class Boss:
             self.run_background_process(cmd, cwd_from=cwd_from, stdin=stdin, env=env)
         return None
 
-    def args_to_special_window(self, args: Iterable[str], cwd_from: Optional[int] = None) -> SpecialWindowInstance:
+    def args_to_special_window(self, args: Iterable[str], cwd_from: Optional[CwdRequest] = None) -> SpecialWindowInstance:
         args = list(args)
         stdin = None
         w = self.active_window
@@ -1781,7 +1799,7 @@ class Boss:
             cmd.append(arg)
         return SpecialWindow(cmd, stdin, cwd_from=cwd_from)
 
-    def _new_tab(self, args: Union[SpecialWindowInstance, Iterable[str]], cwd_from: Optional[int] = None, as_neighbor: bool = False) -> Optional[Tab]:
+    def _new_tab(self, args: Union[SpecialWindowInstance, Iterable[str]], cwd_from: Optional[CwdRequest] = None, as_neighbor: bool = False) -> Optional[Tab]:
         special_window = None
         if args:
             if isinstance(args, SpecialWindowInstance):
@@ -1793,7 +1811,7 @@ class Boss:
             return tm.new_tab(special_window=special_window, cwd_from=cwd_from, as_neighbor=as_neighbor)
         return None
 
-    def _create_tab(self, args: List[str], cwd_from: Optional[int] = None) -> None:
+    def _create_tab(self, args: List[str], cwd_from: Optional[CwdRequest] = None) -> None:
         as_neighbor = False
         if args and args[0].startswith('!'):
             as_neighbor = 'neighbor' in args[0][1:].split(',')
@@ -1806,15 +1824,18 @@ class Boss:
 
     @ac('tab', 'Create a new tab with working directory for the window in it set to the same as the active window')
     def new_tab_with_cwd(self, *args: str) -> None:
-        w = self.active_window_for_cwd
-        cwd_from = w.child.pid_for_cwd if w is not None else None
-        self._create_tab(list(args), cwd_from=cwd_from)
+        self._create_tab(list(args), cwd_from=CwdRequest(self.active_window_for_cwd))
 
-    def new_tab_with_wd(self, wd: str) -> None:
-        special_window = SpecialWindow(None, cwd=wd)
-        self._new_tab(special_window)
+    def new_tab_with_wd(self, wd: Union[str, List[str]], str_is_multiple_paths: bool = False) -> None:
+        if not self.os_window_map:
+            self.add_os_window()
+        if isinstance(wd, str):
+            wd = wd.split(os.pathsep) if str_is_multiple_paths else [wd]
+        for path in wd:
+            special_window = SpecialWindow(None, cwd=path)
+            self._new_tab(special_window)
 
-    def _new_window(self, args: List[str], cwd_from: Optional[int] = None) -> Optional[Window]:
+    def _new_window(self, args: List[str], cwd_from: Optional[CwdRequest] = None) -> Optional[Window]:
         tab = self.active_tab
         if tab is None:
             return None
@@ -1842,8 +1863,7 @@ class Boss:
         w = self.active_window_for_cwd
         if w is None:
             return self.new_window(*args)
-        cwd_from = w.child.pid_for_cwd
-        self._new_window(list(args), cwd_from=cwd_from)
+        self._new_window(list(args), cwd_from=CwdRequest(w))
 
     @ac('misc', '''
         Launch the specified program in a new window/tab/etc.
@@ -1946,8 +1966,10 @@ class Boss:
         ''')
     def load_config_file(self, *paths: str, apply_overrides: bool = True) -> None:
         from .config import load_config
+        from .cli import default_config_paths
         old_opts = get_options()
-        paths = paths or old_opts.config_paths
+        prev_paths = old_opts.config_paths or default_config_paths(self.args.config)
+        paths = paths or prev_paths
         bad_lines: List[BadLine] = []
         opts = load_config(*paths, overrides=old_opts.config_overrides if apply_overrides else None, accumulate_bad_lines=bad_lines)
         if bad_lines:
@@ -2096,18 +2118,19 @@ class Boss:
     def choose_entry(
         self, title: str, entries: Iterable[Tuple[Union[_T, str, None], str]],
         callback: Callable[[Union[_T, str, None]], None],
-        subtitle: str = ''
+        subtitle: str = '',
+        hints_args: Optional[Tuple[str, ...]] = None,
     ) -> Optional[Window]:
-        lines = [title, subtitle, '']
-        idx_map: List[Union[_T, str]] = []
+        lines = [title, subtitle, ' '] if subtitle else [title, ' ']
+        idx_map: List[Union[_T, str, None]] = []
         ans: Union[str, _T, None] = None
         fmt = ': {1}'
 
         for obj, text in entries:
+            idx_map.append(obj)
             if obj is None:
                 lines.append(text)
             else:
-                idx_map.append(obj)
                 lines.append(fmt.format(len(idx_map), text))
 
         def done(data: Dict[str, Any], target_window_id: int, self: Boss) -> None:
@@ -2117,10 +2140,11 @@ class Boss:
         def done2(target_window_id: int, self: Boss) -> None:
             callback(ans)
 
-        q = self._run_kitten(
+        q = self.run_kitten_with_metadata(
             'hints', args=(
                 '--ascending', '--customize-processing=::import::kitty.choose_entry',
-                r'--regex=(?m)^:\s+.+$', '--window-title', title,
+                '--window-title', title,
+                *(hints_args or ())
             ), input_data='\r\n'.join(lines).encode('utf-8'), custom_callback=done, action_on_removal=done2
         )
         return q if isinstance(q, Window) else None
@@ -2136,16 +2160,13 @@ class Boss:
 
         def format_tab_title(tab: Tab) -> str:
             w = 'windows' if tab.num_window_groups > 1 else 'window'
-            return f'{tab.title} [{tab.num_window_groups} {w}]'
+            return f'{tab.name or tab.title} [{tab.num_window_groups} {w}]'
 
         ct = self.active_tab
-        st = ''
-        if ct is not None:
-            st = f'Current tab: {format_tab_title(ct)}'
         self.choose_entry(
             'Choose a tab to switch to',
-            ((t.id, format_tab_title(t)) for t in self.all_tabs if t is not ct),
-            chosen, subtitle=st
+            ((None, f'Current tab: {format_tab_title(t)}') if t is ct else (t.id, format_tab_title(t)) for t in self.all_tabs),
+            chosen
         )
 
     @ac('win', '''
@@ -2208,8 +2229,6 @@ class Boss:
 
     # Can be called with kitty -o "map f1 send_test_notification"
     def send_test_notification(self) -> None:
-        from time import monotonic
-
         from .notify import notify
         now = monotonic()
         ident = f'test-notify-{now}'
@@ -2231,30 +2250,52 @@ class Boss:
             output = '\n'.join(f'{k}={v}' for k, v in os.environ.items())
             self.display_scrollback(w, output, title=_('Current kitty env vars'), report_cursor=False)
 
-    def open_file(self, path: str) -> None:
-        if path == ":cocoa::application launched::":
-            self.cocoa_application_launched = True
-            return
-        from .open_actions import actions_for_launch
+    @ac('debug', '''
+        Close all shared SSH connections
+
+        See :opt:`share_connections <kitten-ssh.share_connections>` for details.
+        ''')
+    def close_shared_ssh_connections(self) -> None:
+        cleanup_ssh_control_masters()
+
+    def launch_urls(self, *urls: str, no_replace_window: bool = False) -> None:
         from .launch import force_window_launch
-        actions = list(actions_for_launch(path))
+        from .open_actions import actions_for_launch
+        actions: List[KeyAction] = []
+        failures = []
+        for url in urls:
+            uactions = tuple(actions_for_launch(url))
+            if uactions:
+                actions.extend(uactions)
+            else:
+                failures.append(url)
         tab = self.active_tab
         if tab is not None:
             w = tab.active_window
         else:
             w = None
-        needs_window_replaced = not self.cocoa_application_launched or not self.os_window_map and w is not None and w.id == 1
+        needs_window_replaced = False
+        if not no_replace_window and not get_options().startup_session:
+            if w is not None and w.id == 1 and monotonic() - w.started_at < 2 and len(tuple(self.all_windows)) == 1:
+                # first window, soon after startup replace it
+                needs_window_replaced = True
 
         def clear_initial_window() -> None:
             if needs_window_replaced and tab is not None and w is not None:
                 tab.remove_window(w)
 
-        if not actions:
-            with force_window_launch(needs_window_replaced):
-                self.launch(kitty_exe(), '+runpy', f'print("The file:", {path!r}, "is of unknown type, cannot open it.");'
-                            'from kitty.utils import hold_till_enter; hold_till_enter(); raise SystemExit(1)')
+        if failures:
+            from kittens.tui.operations import styled
+            spec = '\n  '.join(styled(u, fg='yellow') for u in failures)
+            bdata = json.dumps({'msg': f"Unknown URL type, cannot open:\n  {spec}"}).encode('utf-8')
+            special_window = SpecialWindow([kitty_exe(), '+kitten', 'show_error', '--title', 'Open URL Error'], bdata, 'Open URL Error')
+            if needs_window_replaced and tab is not None:
+                tab.new_special_window(special_window)
+            else:
+                self._new_os_window(special_window)
             clear_initial_window()
-        else:
+            needs_window_replaced = False
+        if actions:
             with force_window_launch(needs_window_replaced):
                 self.dispatch_action(actions.pop(0))
             clear_initial_window()
