@@ -23,6 +23,10 @@
 #include <signal.h>
 extern PyTypeObject Screen_Type;
 
+#if defined(__APPLE__) || defined(__OpenBSD__)
+#define NO_SIGQUEUE 1
+#endif
+
 #ifdef DEBUG_EVENT_LOOP
 #define EVDBG(...) log_event(__VA_ARGS__)
 #else
@@ -133,8 +137,7 @@ new(PyTypeObject *type, PyObject *args, PyObject UNUSED *kwds) {
         return NULL;
     }
     self = (ChildMonitor *)type->tp_alloc(type, 0);
-    if (!init_loop_data(&self->io_loop_data)) return PyErr_SetFromErrno(PyExc_OSError);
-    if (!install_signal_handlers(&self->io_loop_data)) return PyErr_SetFromErrno(PyExc_OSError);
+    if (!init_loop_data(&self->io_loop_data, 5, SIGINT, SIGTERM, SIGCHLD, SIGUSR1, SIGUSR2)) return PyErr_SetFromErrno(PyExc_OSError);
     self->talk_fd = talk_fd;
     self->listen_fd = listen_fd;
     if (self == NULL) return PyErr_NoMemory();
@@ -424,7 +427,7 @@ parse_input(ChildMonitor *self) {
             }
             if (resp) {
                 if (PyBytes_Check(resp)) send_response_to_peer(msg->peer_id, PyBytes_AS_STRING(resp), PyBytes_GET_SIZE(resp));
-                else if (resp == Py_None) send_response_to_peer(msg->peer_id, NULL, 0);
+                else if (resp == Py_None || resp == Py_True) send_response_to_peer(msg->peer_id, NULL, 0);
                 Py_CLEAR(resp);
             } else send_response_to_peer(msg->peer_id, NULL, 0);
         }
@@ -801,11 +804,25 @@ free_twd(ThreadWriteData *x) {
     free(x);
 }
 
+static PyObject*
+sig_queue(PyObject *self UNUSED, PyObject *args) {
+    int pid, signal, value;
+    if (!PyArg_ParseTuple(args, "iii", &pid, &signal, &value)) return NULL;
+#ifdef NO_SIGQUEUE
+    if (kill(pid, signal) != 0) { PyErr_SetFromErrno(PyExc_OSError); return NULL; }
+#else
+    union sigval v;
+    v.sival_int = value;
+    if (sigqueue(pid, signal, v) != 0) { PyErr_SetFromErrno(PyExc_OSError); return NULL; }
+#endif
+    Py_RETURN_NONE;
+}
+
 static PyObject*
 monitor_pid(PyObject *self UNUSED, PyObject *args) {
-    long pid;
+    int pid;
     bool ok = true;
-    if (!PyArg_ParseTuple(args, "l", &pid)) return NULL;
+    if (!PyArg_ParseTuple(args, "i", &pid)) return NULL;
     children_mutex(lock);
     if (monitored_pids_count >= arraysz(monitored_pids)) {
         PyErr_SetString(PyExc_RuntimeError, "Too many monitored pids");
@@ -1224,9 +1241,9 @@ read_bytes(int fd, Screen *screen) {
 typedef struct { bool kill_signal, child_died, reload_config; } SignalSet;
 
 static void
-handle_signal(int signum, void *data) {
+handle_signal(const siginfo_t *siginfo, void *data) {
     SignalSet *ss = data;
-    switch(signum) {
+    switch(siginfo->si_signo) {
         case SIGINT:
         case SIGTERM:
             ss->kill_signal = true;
@@ -1237,6 +1254,9 @@ handle_signal(int signum, void *data) {
         case SIGUSR1:
             ss->reload_config = true;
             break;
+        case SIGUSR2:
+            log_error("Received SIGUSR2: %d\n", siginfo->si_value.sival_int);
+            break;
         default:
             break;
     }
@@ -1416,7 +1436,7 @@ io_loop(void *data) {
 
 typedef struct {
     id_type id;
-    size_t num_of_unresponded_messages_sent_to_main_thread;
+    size_t num_of_unresponded_messages_sent_to_main_thread, fd_array_idx;
     bool finished_reading;
     int fd;
     struct {
@@ -1570,8 +1590,7 @@ wakeup_talk_loop(bool in_signal_handler) {
 
 static void
 prune_peers(void) {
-    for (size_t i = 0; i < talk_data.num_peers; i++) {
-        size_t idx = talk_data.num_peers - 1 - i;
+    for (size_t idx = talk_data.num_peers; idx-- > 0;) {
         Peer *p = talk_data.peers + idx;
         if (p->read.finished && !p->num_of_unresponded_messages_sent_to_main_thread && !p->write.used) {
             free_peer(p);
@@ -1585,7 +1604,7 @@ talk_loop(void *data) {
     // The talk thread loop
     ChildMonitor *self = (ChildMonitor*)data;
     set_thread_name("KittyPeerMon");
-    if (!init_loop_data(&talk_data.loop_data)) { log_error("Failed to create wakeup fd for talk thread with error: %s", strerror(errno)); }
+    if (!init_loop_data(&talk_data.loop_data, 0)) { log_error("Failed to create wakeup fd for talk thread with error: %s", strerror(errno)); }
     PollFD fds[PEER_LIMIT + 8] = {{0}};
     size_t num_listen_fds = 0, num_peer_fds = 0;
 #define add_listener(which) \
@@ -1604,13 +1623,14 @@ talk_loop(void *data) {
             for (size_t i = 0; i < talk_data.num_peers; i++) {
                 Peer *p = talk_data.peers + i;
                 if (!p->read.finished || p->write.used) {
-                    fds[num_listen_fds + num_peer_fds].fd = p->fd;
-                    fds[num_listen_fds + num_peer_fds].revents = 0;
+                    p->fd_array_idx = num_listen_fds + num_peer_fds++;
+                    fds[p->fd_array_idx].fd = p->fd;
+                    fds[p->fd_array_idx].revents = 0;
                     int flags = 0;
                     if (!p->read.finished) flags |= POLLIN;
                     if (p->write.used) flags |= POLLOUT;
-                    fds[num_listen_fds + num_peer_fds++].events = flags;
-                }
+                    fds[p->fd_array_idx].events = flags;
+                } else p->fd_array_idx = 0;
             }
             talk_mutex(unlock);
         }
@@ -1625,13 +1645,16 @@ talk_loop(void *data) {
             if (fds[num_listen_fds - 1].revents & POLLIN) {
                 drain_fd(fds[num_listen_fds - 1].fd);  // wakeup
             }
-            for (size_t i = num_listen_fds, k = 0; i < num_peer_fds + num_listen_fds; i++, k++) {
+            for (size_t k = 0; k < talk_data.num_peers; k++) {
                 Peer *p = talk_data.peers + k;
-                if (fds[i].revents & (POLLIN | POLLHUP)) read_from_peer(self, p);
-                if (fds[i].revents & POLLOUT) write_to_peer(p);
-                if (fds[i].revents & POLLNVAL) {
-                    p->read.finished = true;
-                    p->write.failed = true; p->write.used = 0;
+                if (p->fd_array_idx) {
+                    if (fds[p->fd_array_idx].revents & (POLLIN | POLLHUP)) read_from_peer(self, p);
+                    if (fds[p->fd_array_idx].revents & POLLOUT) write_to_peer(p);
+                    if (fds[p->fd_array_idx].revents & POLLNVAL) {
+                        p->read.finished = true;
+                        p->write.failed = true; p->write.used = 0;
+                    }
+                    break;
                 }
             }
         } else if (ret < 0) { if (errno != EAGAIN && errno != EINTR) perror("poll() on talk fds failed"); }
@@ -1736,6 +1759,7 @@ static PyMethodDef module_methods[] = {
     METHODB(monitor_pid, METH_VARARGS),
     METHODB(send_data_to_peer, METH_VARARGS),
     METHODB(cocoa_set_menubar_title, METH_VARARGS),
+    {"sigqueue", (PyCFunction)sig_queue, METH_VARARGS, ""},
     {NULL}  /* Sentinel */
 };
 
@@ -1745,6 +1769,11 @@ init_child_monitor(PyObject *module) {
     if (PyModule_AddObject(module, "ChildMonitor", (PyObject *)&ChildMonitor_Type) != 0) return false;
     Py_INCREF(&ChildMonitor_Type);
     if (PyModule_AddFunctions(module, module_methods) != 0) return false;
+#ifdef NO_SIGQUEUE
+    PyModule_AddIntConstant(module, "has_sigqueue", 0);
+#else
+    PyModule_AddIntConstant(module, "has_sigqueue", 1);
+#endif
     return true;
 }
 
