@@ -270,7 +270,7 @@ def zsh_output_serializer(ans: Completions) -> str:
         yield ans
 
     for description, matches in ans.match_groups.items():
-        cmd = ['compadd', '-U', '-J', shlex.quote(description), '-X', shlex.quote('%B' + description + '%b')]
+        cmd = ['compadd', '-U', '-J', shlex.quote(description), '-X', shlex.quote(f'%B{description}%b')]
         if not matches.trailing_space:
             cmd += ['-S', '""']
         if matches.is_files:
@@ -285,7 +285,10 @@ def zsh_output_serializer(ans: Completions) -> str:
         has_descriptions = any(matches.values())
         if has_descriptions or matches.word_transforms:
             lines.append('compdescriptions=(')
-            sz = max(map(wcswidth, matches.transformed_words()))
+            try:
+                sz = max(map(wcswidth, matches.transformed_words()))
+            except ValueError:
+                sz = 0
             limit = min(16, sz)
             for word, desc in matches.transformed_items():
                 lines.extend(map(shlex.quote, fmt_desc(word, desc, limit)))
@@ -350,10 +353,10 @@ def fish2_output_serializer(ans: Completions) -> str:
 
 
 def completions_for_first_word(ans: Completions, prefix: str, entry_points: Iterable[str], namespaced_entry_points: Iterable[str]) -> None:
-    cmds = ['@' + c for c in remote_control_command_names()]
+    cmds = [f'@{c}' for c in remote_control_command_names()]
     ans.add_match_group('Entry points', {
         k: '' for k in
-        list(entry_points) + cmds + ['+' + k for k in namespaced_entry_points]
+        list(entry_points) + cmds + [f'+{k}' for k in namespaced_entry_points]
         if not prefix or k.startswith(prefix)
     })
     if prefix:
@@ -443,7 +446,7 @@ def complete_alias_map(
                 long_opt = option_map.get(parts[0])
                 if long_opt is not None and complete_args is not None:
                     complete_args(ans, long_opt, parts[1], Delegate())
-                    ans.add_prefix(parts[0] + '=')
+                    ans.add_prefix(f'{parts[0]}=')
                 return
         opt = option_map.get(w)
         if w is last_word and not new_word:
@@ -502,6 +505,16 @@ def complete_remote_command(ans: Completions, cmd_name: str, words: Sequence[str
     complete_alias_map(ans, words, new_word, alias_map, complete_args=args_completer)
 
 
+def complete_launch_wrapper(ans: Completions, words: Sequence[str], new_word: bool, allow_files: bool = True) -> None:
+    from kitty.launch import clone_safe_opts
+    aliases, alias_map = options_for_cmd('launch')
+    alias_map = {k: v for k, v in alias_map.items() if v['dest'] in clone_safe_opts()}
+    args_completer: Optional[CompleteArgsFunc] = None
+    if allow_files:
+        args_completer = remote_files_completer('Files', ('*',))
+    complete_alias_map(ans, words, new_word, alias_map, complete_args=args_completer)
+
+
 def path_completion(prefix: str = '') -> Tuple[List[str], List[str]]:
     prefix = prefix.replace(r'\ ', ' ')
     dirs, files = [], []
@@ -661,7 +674,12 @@ def complete_kitten(ans: Completions, kitten: str, words: Sequence[str], new_wor
 
 def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterable[str], namespaced_entry_points: Iterable[str]) -> Completions:
     ans = Completions()
-    if not words or words[0] != 'kitty':
+    if not words:
+        return ans
+    if words[0] in ('edit-in-kitty', 'clone-in-kitty'):
+        complete_launch_wrapper(ans, words[1:], new_word, allow_files=words[0] != 'clone-in-kitty')
+        return ans
+    if words[0] != 'kitty':
         return ans
     words = words[1:]
     if not words or (len(words) == 1 and not new_word):
@@ -682,7 +700,7 @@ def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterabl
     if words[0].startswith('@'):
         if len(words) == 1 and not new_word:
             prefix = words[0]
-            ans.add_match_group('Remote control commands', {'@' + c: '' for c in remote_control_command_names() if c.startswith(prefix)})
+            ans.add_match_group('Remote control commands', {f'@{c}': '' for c in remote_control_command_names() if c.startswith(prefix)})
         else:
             complete_remote_command(ans, words[0][1:], words[1:], new_word)
     if words[0] == '+':
@@ -695,21 +713,27 @@ def find_completions(words: Sequence[str], new_word: bool, entry_points: Iterabl
                     ans.add_match_group('Kittens', (k for k in all_kitten_names() if k.startswith('' if len(words) == 2 else words[2])))
                 else:
                     complete_kitten(ans, words[2], words[3:], new_word)
+            elif words[1] == 'open':
+                complete_cli(ans, words[2:], new_word, options_for_completion(), complete_kitty_cli_arg)
         return ans
     if words[0].startswith('+'):
         if len(words) == 1:
             if new_word:
                 if words[0] == '+kitten':
                     ans.add_match_group('Kittens', all_kitten_names())
+                elif words[0] == '+open':
+                    complete_cli(ans, words[1:], new_word, options_for_completion(), complete_kitty_cli_arg)
             else:
                 prefix = words[0]
                 ans.add_match_group('Entry points', (c for c in namespaced_entry_points if c.startswith(prefix)))
         else:
-            if len(words) == 2 and not new_word:
-                ans.add_match_group('Kittens', (k for k in all_kitten_names() if k.startswith(words[1])))
-            else:
-                if words[0] == '+kitten':
+            if words[0] == '+kitten':
+                if len(words) == 2 and not new_word:
+                    ans.add_match_group('Kittens', (k for k in all_kitten_names() if k.startswith(words[1])))
+                else:
                     complete_kitten(ans, words[1], words[2:], new_word)
+            elif words[0] == '+open':
+                complete_cli(ans, words[1:], new_word, options_for_completion(), complete_kitty_cli_arg)
     else:
         complete_cli(ans, words, new_word, options_for_completion(), complete_kitty_cli_arg)
 
