@@ -22,16 +22,38 @@ class Version(NamedTuple):
 
 appname: str = 'kitty'
 kitty_face = '🐱'
-version: Version = Version(0, 24, 2)
+version: Version = Version(0, 25, 2)
 str_version: str = '.'.join(map(str, version))
 _plat = sys.platform.lower()
 is_macos: bool = 'darwin' in _plat
+is_freebsd: bool = 'freebsd' in _plat
+is_running_from_develop: bool = False
 if getattr(sys, 'frozen', False):
     extensions_dir: str = getattr(sys, 'kitty_extensions_dir')
-    kitty_base_dir = os.path.dirname(extensions_dir)
-    if is_macos:
-        kitty_base_dir = os.path.dirname(os.path.dirname(kitty_base_dir))
-    kitty_base_dir = os.path.join(kitty_base_dir, 'kitty')
+
+    def get_frozen_base() -> str:
+        global is_running_from_develop
+        try:
+            from bypy_importer import running_in_develop_mode  # type: ignore
+        except ImportError:
+            pass
+        else:
+            is_running_from_develop = running_in_develop_mode()
+
+        if is_running_from_develop:
+            q = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
+            try:
+                if os.path.isdir(q):
+                    return q
+            except OSError:
+                pass
+        ans = os.path.dirname(extensions_dir)
+        if is_macos:
+            ans = os.path.dirname(os.path.dirname(ans))
+        ans = os.path.join(ans, 'kitty')
+        return ans
+    kitty_base_dir = get_frozen_base()
+    del get_frozen_base
 else:
     kitty_base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
     extensions_dir = os.path.join(kitty_base_dir, 'kitty')
@@ -116,6 +138,26 @@ def cache_dir() -> str:
     return candidate
 
 
+@run_once
+def runtime_dir() -> str:
+    if 'KITTY_RUNTIME_DIRECTORY' in os.environ:
+        candidate = os.path.abspath(os.environ['KITTY_RUNTIME_DIRECTORY'])
+    elif is_macos:
+        from .fast_data_types import user_cache_dir
+        candidate = user_cache_dir()
+    elif 'XDG_RUNTIME_DIR' in os.environ:
+        candidate = os.path.abspath(os.environ['XDG_RUNTIME_DIR'])
+    else:
+        candidate = f'/run/user/{os.geteuid()}'
+        if not os.path.isdir(candidate) or not os.access(candidate, os.X_OK | os.W_OK | os.R_OK):
+            candidate = os.path.join(cache_dir(), 'run')
+    os.makedirs(candidate, exist_ok=True)
+    import stat
+    if stat.S_IMODE(os.stat(candidate).st_mode) != 0o700:
+        os.chmod(candidate, 0o700)
+    return candidate
+
+
 def wakeup() -> None:
     from .fast_data_types import get_boss
     b = get_boss()
@@ -133,6 +175,9 @@ except KeyError:
     with suppress(Exception):
         print('Failed to read login shell via getpwuid() for current user, falling back to /bin/sh', file=sys.stderr)
     shell_path = '/bin/sh'
+# Keep this short as it is limited to 103 bytes on macOS
+# https://github.com/ansible/ansible/issues/11536#issuecomment-153030743
+ssh_control_master_template = 'kssh-{kitty_pid}-{ssh_placeholder}'
 
 
 def glfw_path(module: str) -> str:
