@@ -928,8 +928,14 @@ group_iosevka(Font *font, hb_font_t *hbf) {
         const LigatureType current = ligature_types[G(glyph_idx)];
         const LigatureType after = is_last_glyph ? LIGATURE_UNKNOWN : ligature_types[G(glyph_idx + 1)];
         bool end_current_group = false;
-        if (current_group->num_glyphs && is_iosevka_lig_ender(before, current, after)) {
-            end_current_group = true;
+        if (current_group->num_glyphs) {
+            if (is_iosevka_lig_ender(before, current, after)) end_current_group = true;
+            else {
+                if (!current_group->num_cells && !current_group->has_special_glyph) {
+                    if (is_iosevka_lig_starter(before, current, after)) current_group->has_special_glyph = true;
+                    else end_current_group = true;
+                }
+            }
         }
         if (!current_group->num_glyphs++) {
             if (is_iosevka_lig_starter(before, current, after)) current_group->has_special_glyph = true;
@@ -954,7 +960,7 @@ group_iosevka(Font *font, hb_font_t *hbf) {
             }
             current_group->num_cells += num_cells_consumed;
         }
-        if (end_current_group) G(group_idx)++;
+        if (end_current_group && current_group->num_cells) G(group_idx)++;
         G(glyph_idx)++;
     }
 }
@@ -1017,9 +1023,9 @@ group_normal(Font *font, hb_font_t *hbf) {
 #if 0
         char ch[8] = {0};
         encode_utf8(G(current_cell_data).current_codepoint, ch);
-        printf("\x1b[32m→ %s\x1b[m glyph_idx: %zu glyph_id: %u (%s) group_idx: %zu cluster: %u -> %u is_special: %d\n"
+        printf("\x1b[32m→ %s\x1b[m glyph_idx: %zu glyph_id: %u group_idx: %zu cluster: %u -> %u is_special: %d\n"
                 "  num_codepoints_used_by_glyph: %u current_group: (%u cells, %u glyphs) add_to_current_group: %d\n",
-                ch, G(glyph_idx), glyph_id, glyph_name, G(group_idx), cluster, next_cluster, is_special,
+                ch, G(glyph_idx), glyph_id, G(group_idx), cluster, next_cluster, is_special,
                 num_codepoints_used_by_glyph, current_group->num_cells, current_group->num_glyphs, add_to_current_group);
 #endif
 
