@@ -25,6 +25,7 @@ extern void cocoa_set_titlebar_appearance(void *w, unsigned int theme);
 extern void cocoa_set_titlebar_color(void *w, color_type color);
 extern bool cocoa_alt_option_key_pressed(unsigned long);
 extern void cocoa_toggle_secure_keyboard_entry(void);
+extern void cocoa_update_menu_bar_title(PyObject*);
 extern size_t cocoa_get_workspace_ids(void *w, size_t *workspace_ids, size_t array_sz);
 extern monotonic_t cocoa_cursor_blink_interval(void);
 
@@ -46,6 +47,43 @@ get_platform_dependent_config_values(void *glfw_window) {
     }
 }
 
+static void
+strip_csi_(const char *title, char *buf, size_t bufsz) {
+    enum { NORMAL, IN_ESC, IN_CSI} state = NORMAL;
+    char *dest = buf, *last = &buf[bufsz-1];
+    *dest = 0; *last = 0;
+
+    for (; *title && dest < last; title++) {
+        const char ch = *title;
+        switch (state) {
+            case NORMAL: {
+                if (ch == 0x1b) { state = IN_ESC; }
+                else *(dest++) = ch;
+            } break;
+            case IN_ESC: {
+                if (ch == '[') { state = IN_CSI; }
+                else { state = NORMAL; }
+            } break;
+            case IN_CSI: {
+                if (!(('0' <= ch && ch <= '9') || ch == ';' || ch == ':')) state = NORMAL;
+            } break;
+        }
+    }
+    *dest = 0;
+}
+
+
+void
+update_menu_bar_title(PyObject *title UNUSED) {
+#ifdef __APPLE__
+    static char buf[2048];
+    strip_csi_(PyUnicode_AsUTF8(title), buf, arraysz(buf));
+    DECREF_AFTER_FUNCTION PyObject *stitle = PyUnicode_FromString(buf);
+    cocoa_update_menu_bar_title(stitle);
+#endif
+}
+
+
 void
 request_tick_callback(void) {
     glfwPostEmptyEvent();
@@ -429,8 +467,8 @@ static void get_window_dpi(GLFWwindow *w, double *x, double *y);
 
 #ifdef __APPLE__
 static bool
-apple_file_open_callback(const char* filepath) {
-    set_cocoa_pending_action(OPEN_FILE, filepath);
+apple_url_open_callback(const char* url) {
+    set_cocoa_pending_action(LAUNCH_URLS, url);
     return true;
 }
 
@@ -438,7 +476,9 @@ extern bool cocoa_render_line_of_text(const char *text, const color_type fg, con
 
 bool
 draw_window_title(OSWindow *window UNUSED, const char *text, color_type fg, color_type bg, uint8_t *output_buf, size_t width, size_t height) {
-    return cocoa_render_line_of_text(text, fg, bg, output_buf, width, height);
+    static char buf[2048];
+    strip_csi_(text, buf, arraysz(buf));
+    return cocoa_render_line_of_text(buf, fg, bg, output_buf, width, height);
 }
 
 extern uint8_t* render_single_ascii_char_as_mask(const char ch, size_t *result_width, size_t *result_height);
@@ -484,10 +524,12 @@ draw_text_callback(GLFWwindow *window, const char *text, uint32_t fg, uint32_t b
 bool
 draw_window_title(OSWindow *window, const char *text, color_type fg, color_type bg, uint8_t *output_buf, size_t width, size_t height) {
     if (!ensure_csd_title_render_ctx()) return false;
+    static char buf[2048];
+    strip_csi_(text, buf, arraysz(buf));
     unsigned px_sz = (unsigned)(window->fonts_data->font_sz_in_pts * window->fonts_data->logical_dpi_y / 72.);
     px_sz = MIN(px_sz, 3 * height / 4);
 #define RGB2BGR(x) (x & 0xFF000000) | ((x & 0xFF0000) >> 16) | (x & 0x00FF00) | ((x & 0x0000FF) << 16)
-    bool ok = render_single_line(csd_title_render_ctx, text, px_sz, RGB2BGR(fg), RGB2BGR(bg), output_buf, width, height, 0, 0, 0);
+    bool ok = render_single_line(csd_title_render_ctx, buf, px_sz, RGB2BGR(fg), RGB2BGR(bg), output_buf, width, height, 0, 0, 0);
 #undef RGB2BGR
     if (!ok && PyErr_Occurred()) PyErr_Print();
     return ok;
@@ -733,6 +775,7 @@ create_os_window(PyObject UNUSED *self, PyObject *args, PyObject *kw) {
     glfwWindowHintString(GLFW_X11_INSTANCE_NAME, wm_class_name);
     glfwWindowHintString(GLFW_X11_CLASS_NAME, wm_class_class);
     glfwWindowHintString(GLFW_WAYLAND_APP_ID, wm_class_class);
+    glfwWindowHint(GLFW_COCOA_COLOR_SPACE, OPT(macos_colorspace));
 #endif
 
     if (global_state.num_os_windows >= MAX_CHILDREN) {
@@ -1003,7 +1046,7 @@ glfw_init(PyObject UNUSED *self, PyObject *args) {
     PyObject *ans = glfwInit(monotonic_start_time) ? Py_True: Py_False;
     if (ans == Py_True) {
 #ifdef __APPLE__
-        glfwSetCocoaFileOpenCallback(apple_file_open_callback);
+        glfwSetCocoaURLOpenCallback(apple_url_open_callback);
 #else
         glfwSetDrawTextFunction(draw_text_callback);
 #endif
@@ -1286,31 +1329,6 @@ request_window_attention(id_type kitty_window_id, bool audio_bell) {
     }
 }
 
-static void
-strip_csi_(const char *title, char *buf, size_t bufsz) {
-    enum { NORMAL, IN_ESC, IN_CSI} state = NORMAL;
-    char *dest = buf, *last = &buf[bufsz-1];
-    *dest = 0; *last = 0;
-
-    for (; *title && dest < last; title++) {
-        const char ch = *title;
-        switch (state) {
-            case NORMAL: {
-                if (ch == 0x1b) { state = IN_ESC; }
-                else *(dest++) = ch;
-            } break;
-            case IN_ESC: {
-                if (ch == '[') { state = IN_CSI; }
-                else { state = NORMAL; }
-            } break;
-            case IN_CSI: {
-                if (!(('0' <= ch && ch <= '9') || ch == ';' || ch == ':')) state = NORMAL;
-            } break;
-        }
-    }
-    *dest = 0;
-}
-
 void
 set_os_window_title(OSWindow *w, const char *title) {
     static char buf[2048];
