@@ -42,12 +42,16 @@ typedef struct {
     bool rectangle_select, adjusting_start;
     IterationData last_rendered;
     int sort_y, sort_x;
+    struct {
+        SelectionBoundary start, end;
+        unsigned int scrolled_by;
+    } initial_extent;
 } Selection;
 
 typedef struct {
     Selection *items;
     size_t count, capacity, last_rendered_count;
-    bool in_progress;
+    bool in_progress, extension_in_progress;
     SelectionExtendMode extend_mode;
 } Selections;
 
@@ -63,11 +67,6 @@ typedef struct {
 } Savepoint;
 
 
-typedef struct {
-    ScreenModes buf[SAVEPOINTS_SZ];
-    index_type start_of_data, count;
-} SavemodesBuffer;
-
 typedef struct {
     CPUCell *cpu_cells;
     GPUCell *gpu_cells;
@@ -94,14 +93,13 @@ typedef struct {
     bool use_latin1, is_dirty, scroll_changed, reload_all_gpu_data;
     Cursor *cursor;
     Savepoint main_savepoint, alt_savepoint;
-    SavemodesBuffer modes_savepoints;
     PyObject *callbacks, *test_child;
     LineBuf *linebuf, *main_linebuf, *alt_linebuf;
     GraphicsManager *grman, *main_grman, *alt_grman;
     HistoryBuf *historybuf;
     unsigned int history_line_added_count;
     bool *tabstops, *main_tabstops, *alt_tabstops;
-    ScreenModes modes;
+    ScreenModes modes, saved_modes;
     ColorProfile *color_profile;
     monotonic_t start_visual_bell_at;
 
@@ -114,6 +112,7 @@ typedef struct {
     pthread_mutex_t read_buf_lock, write_buf_lock;
 
     CursorRenderInfo cursor_render_info;
+    unsigned int render_unfocused_cursor;
 
     struct {
         size_t capacity, used;
@@ -130,6 +129,27 @@ typedef struct {
     ANSIBuf as_ansi_buf;
     char_type last_graphic_char;
     uint8_t main_key_encoding_flags[8], alt_key_encoding_flags[8], *key_encoding_flags;
+    struct {
+        monotonic_t start, duration;
+    } ignore_bells;
+    union {
+        struct {
+            unsigned int redraws_prompts_at_all: 1;
+        };
+        unsigned int val;
+    } prompt_settings;
+    char display_window_char;
+    struct {
+        char ch;
+        uint8_t *canvas;
+        size_t requested_height, width_px, height_px;
+    } last_rendered_window_char;
+    struct {
+        unsigned int scrolled_by;
+        index_type y;
+        bool is_set;
+    } last_visited_prompt;
+    PyObject *last_reported_cwd;
 } Screen;
 
 
@@ -139,8 +159,10 @@ void screen_align(Screen*);
 void screen_restore_cursor(Screen *);
 void screen_save_cursor(Screen *);
 void screen_restore_modes(Screen *);
+void screen_restore_mode(Screen *, unsigned int);
 void screen_save_modes(Screen *);
-void write_escape_code_to_child(Screen *self, unsigned char which, const char *data);
+void screen_save_mode(Screen *, unsigned int);
+bool write_escape_code_to_child(Screen *self, unsigned char which, const char *data);
 void screen_cursor_position(Screen*, unsigned int, unsigned int);
 void screen_cursor_back(Screen *self, unsigned int count/*=1*/, int move_direction/*=-1*/);
 void screen_erase_in_line(Screen *, unsigned int, bool);
@@ -186,7 +208,7 @@ void screen_handle_cmd(Screen *, PyObject *cmd);
 void screen_push_colors(Screen *, unsigned int);
 void screen_pop_colors(Screen *, unsigned int);
 void screen_report_color_stack(Screen *);
-void screen_handle_print(Screen *, PyObject *cmd);
+void screen_handle_kitty_dcs(Screen *, const char *callback_name, PyObject *cmd);
 void screen_designate_charset(Screen *, uint32_t which, uint32_t as);
 void screen_use_latin1(Screen *, bool);
 void set_title(Screen *self, PyObject*);
@@ -194,6 +216,8 @@ void desktop_notify(Screen *self, unsigned int, PyObject*);
 void set_icon(Screen *self, PyObject*);
 void set_dynamic_color(Screen *self, unsigned int code, PyObject*);
 void clipboard_control(Screen *self, int code, PyObject*);
+void shell_prompt_marking(Screen *self, PyObject*);
+void file_transmission(Screen *self, PyObject*);
 void set_color_table_color(Screen *self, unsigned int code, PyObject*);
 void process_cwd_notification(Screen *self, unsigned int code, PyObject*);
 uint32_t* translation_table(uint32_t which);
@@ -208,7 +232,7 @@ bool screen_is_selection_dirty(Screen *self);
 bool screen_has_selection(Screen*);
 bool screen_invert_colors(Screen *self);
 void screen_update_cell_data(Screen *self, void *address, FONTS_DATA_HANDLE, bool cursor_has_moved);
-bool screen_is_cursor_visible(Screen *self);
+bool screen_is_cursor_visible(const Screen *self);
 bool screen_selection_range_for_line(Screen *self, index_type y, index_type *start, index_type *end);
 bool screen_selection_range_for_word(Screen *self, const index_type x, const index_type y, index_type *, index_type *, index_type *start, index_type *end, bool);
 void screen_start_selection(Screen *self, index_type x, index_type y, bool, bool, SelectionExtendMode);
@@ -224,6 +248,8 @@ void set_active_hyperlink(Screen*, char*, char*);
 hyperlink_id_type screen_mark_hyperlink(Screen*, index_type, index_type);
 void screen_handle_graphics_command(Screen *self, const GraphicsCommand *cmd, const uint8_t *payload);
 bool screen_open_url(Screen*);
+bool screen_set_last_visited_prompt(Screen*, index_type);
+bool screen_select_cmd_output(Screen*, index_type);
 void screen_dirty_sprite_positions(Screen *self);
 void screen_rescale_images(Screen *self);
 void screen_report_size(Screen *, unsigned int which);
@@ -234,8 +260,9 @@ void screen_push_key_encoding_flags(Screen *self, uint32_t val);
 void screen_pop_key_encoding_flags(Screen *self, uint32_t num);
 uint8_t screen_current_key_encoding_flags(Screen *self);
 void screen_report_key_encoding_flags(Screen *self);
-void screen_xtmodkeys(Screen *self, uint32_t p1, uint32_t p2);
 bool screen_detect_url(Screen *screen, unsigned int x, unsigned int y);
+int screen_cursor_at_a_shell_prompt(const Screen *);
+bool screen_fake_move_cursor_to_position(Screen *, index_type x, index_type y);
 #define DECLARE_CH_SCREEN_HANDLER(name) void screen_##name(Screen *screen);
 DECLARE_CH_SCREEN_HANDLER(bell)
 DECLARE_CH_SCREEN_HANDLER(backspace)
