@@ -3,13 +3,16 @@
 
 import json
 import os
+import re
 import sys
 import weakref
 from collections import deque
-from enum import IntEnum
-from functools import partial
+from contextlib import suppress
+from enum import Enum, IntEnum, auto
+from functools import lru_cache, partial
 from gettext import gettext as _
 from itertools import chain
+from time import monotonic
 from typing import (
     TYPE_CHECKING, Any, Callable, Deque, Dict, Iterable, List, NamedTuple,
     Optional, Pattern, Sequence, Tuple, Union
@@ -18,16 +21,17 @@ from typing import (
 from .child import ProcessDesc
 from .cli_stub import CLIOptions
 from .config import build_ansi_color_table
-from .constants import appname, is_macos, wakeup
+from .constants import appname, config_dir, is_macos, wakeup
 from .fast_data_types import (
     BGIMAGE_PROGRAM, BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM,
     CELL_PROGRAM, CELL_SPECIAL_PROGRAM, CURSOR_BEAM, CURSOR_BLOCK,
-    CURSOR_UNDERLINE, DCS, DECORATION, DIM, GLFW_MOD_CONTROL,
+    CURSOR_UNDERLINE, DCS, DECORATION, DECORATION_MASK, DIM, GLFW_MOD_CONTROL,
     GRAPHICS_ALPHA_MASK_PROGRAM, GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM,
-    MARK, MARK_MASK, NO_CURSOR_SHAPE, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE,
-    SCROLL_PAGE, STRIKETHROUGH, TINT_PROGRAM, Color, KeyEvent, Screen,
-    add_timer, add_window, cell_size_for_window, click_mouse_cmd_output,
-    click_mouse_url, compile_program, encode_key_for_tty, get_boss,
+    MARK, MARK_MASK, NO_CURSOR_SHAPE, NUM_UNDERLINE_STYLES, OSC, REVERSE,
+    SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE, STRIKETHROUGH, TINT_PROGRAM, Color,
+    KeyEvent, Screen, add_timer, add_window, cell_size_for_window,
+    click_mouse_cmd_output, click_mouse_url, compile_program,
+    current_os_window, encode_key_for_tty, get_boss, get_click_interval,
     get_clipboard_string, get_options, init_cell_program, mark_os_window_dirty,
     mouse_selection, move_cursor_to_mouse_if_in_prompt, pt_to_px,
     set_clipboard_string, set_titlebar_color, set_window_logo,
@@ -39,12 +43,12 @@ from .notify import NotificationCommand, handle_notification_cmd
 from .options.types import Options
 from .rgb import to_color
 from .terminfo import get_capabilities
-from .types import MouseEvent, WindowGeometry, ac
+from .types import MouseEvent, WindowGeometry, ac, run_once
 from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
-    get_primary_selection, load_shaders, log_error, open_cmd, open_url,
-    parse_color_set, resolve_custom_file, sanitize_title, set_primary_selection,
-    sgr_sanitizer_pat
+    get_primary_selection, kitty_ansi_sanitizer_pat, load_shaders, log_error,
+    open_cmd, open_url, parse_color_set, path_from_osc7_url,
+    resolve_custom_file, resolved_shell, sanitize_title, set_primary_selection
 )
 
 MatchPatternType = Union[Pattern[str], Tuple[Pattern[str], Optional[Pattern[str]]]]
@@ -54,6 +58,77 @@ if TYPE_CHECKING:
     from .file_transmission import FileTransmission
 
 
+class CwdRequestType(Enum):
+    current: int = auto()
+    last_reported: int = auto()
+    oldest: int = auto()
+
+
+class CwdRequest:
+
+    def __init__(self, window: Optional['Window'] = None, request_type: CwdRequestType = CwdRequestType.current) -> None:
+        self.window_id = -1 if window is None else window.id
+        self.request_type = request_type
+
+    def __bool__(self) -> bool:
+        return self.window_id > -1
+
+    @property
+    def window(self) -> Optional['Window']:
+        return get_boss().window_id_map.get(self.window_id)
+
+    @property
+    def cwd_of_child(self) -> str:
+        window = self.window
+        if not window:
+            return ''
+        reported_cwd = path_from_osc7_url(window.screen.last_reported_cwd) if window.screen.last_reported_cwd else ''
+        if reported_cwd and not window.child_is_remote and (self.request_type is CwdRequestType.last_reported or window.at_prompt):
+            return reported_cwd
+        return window.get_cwd_of_child(oldest=self.request_type is CwdRequestType.oldest) or ''
+
+    def modify_argv_for_launch_with_cwd(self, argv: List[str]) -> str:
+        window = self.window
+        if not window:
+            return ''
+        reported_cwd = path_from_osc7_url(window.screen.last_reported_cwd) if window.screen.last_reported_cwd else ''
+        if reported_cwd:
+            # First check if we are running ssh kitten, and trying to open the configured login shell
+            if argv[0] == resolved_shell(get_options())[0]:
+                ssh_kitten_cmdline = window.ssh_kitten_cmdline()
+                if ssh_kitten_cmdline:
+                    from kittens.ssh.utils import set_cwd_in_cmdline
+                    argv[:] = ssh_kitten_cmdline
+                    set_cwd_in_cmdline(reported_cwd, argv)
+                    return ''
+            if not window.child_is_remote and (self.request_type is CwdRequestType.last_reported or window.at_prompt):
+                return reported_cwd
+        return window.get_cwd_of_child(oldest=self.request_type is CwdRequestType.oldest) or ''
+
+
+def process_title_from_child(title: str, is_base64: bool) -> str:
+    if is_base64:
+        from base64 import standard_b64decode
+        try:
+            title = standard_b64decode(title).decode('utf-8', 'replace')
+        except Exception:
+            title = 'undecodeable title'
+    return sanitize_title(title)
+
+
+@lru_cache(maxsize=64)
+def compile_match_query(exp: str, is_simple: bool = True) -> MatchPatternType:
+    if is_simple:
+        pat: MatchPatternType = re.compile(exp)
+    else:
+        kp, vp = exp.partition('=')[::2]
+        if vp:
+            pat = re.compile(kp), re.compile(vp)
+        else:
+            pat = re.compile(kp), None
+    return pat
+
+
 class WindowDict(TypedDict):
     id: int
     is_focused: bool
@@ -89,7 +164,7 @@ class DynamicColor(IntEnum):
 
 
 class CommandOutput(IntEnum):
-    last_run, first_on_screen, last_visited = 0, 1, 2
+    last_run, first_on_screen, last_visited, last_non_empty = 0, 1, 2, 3
 
 
 DYNAMIC_COLOR_CODES = {
@@ -154,6 +229,14 @@ def call_watchers(windowref: Callable[[], Optional['Window']], which: str, data:
     add_timer(callback, 0, False)
 
 
+def pagerhist(screen: Screen, as_ansi: bool = False, add_wrap_markers: bool = True, upto_output_start: bool = False) -> str:
+    pht = screen.historybuf.pagerhist_as_text(upto_output_start)
+    if pht and (not as_ansi or not add_wrap_markers):
+        sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
+        pht = sanitizer(pht)
+    return pht
+
+
 def as_text(
     screen: Screen,
     as_ansi: bool = False,
@@ -183,13 +266,8 @@ def as_text(
             ctext += f'\x1b[{code} q'
 
     if add_history:
-        h: List[str] = []
-        pht = screen.historybuf.pagerhist_as_text()
-        if pht:
-            h.append(pht)
-        if h and (not as_ansi or not add_wrap_markers):
-            sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
-            h = list(map(sanitizer, h))
+        pht = pagerhist(screen, as_ansi, add_wrap_markers)
+        h: List[str] = [pht] if pht else []
         screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
         if h:
             if not screen.linebuf.is_continued(0):
@@ -206,6 +284,15 @@ def as_text(
     return ans
 
 
+def multi_replace(src: str, **replacements: Any) -> str:
+    r = {k: str(v) for k, v in replacements.items()}
+
+    def sub(m: 're.Match[str]') -> str:
+        return r.get(m.group(1), m.group(1))
+
+    return re.sub(r'\{([A-Z_]+)\}', sub, src)
+
+
 class LoadShaderPrograms:
 
     def __call__(self, semi_transparent: bool = False) -> None:
@@ -218,16 +305,19 @@ class LoadShaderPrograms:
                 'SPECIAL': CELL_SPECIAL_PROGRAM,
                 'FOREGROUND': CELL_FG_PROGRAM,
         }.items():
-            vv, ff = v.replace('WHICH_PROGRAM', which), f.replace('WHICH_PROGRAM', which)
-            for gln, pyn in {
-                    'REVERSE_SHIFT': REVERSE,
-                    'STRIKE_SHIFT': STRIKETHROUGH,
-                    'DIM_SHIFT': DIM,
-                    'DECORATION_SHIFT': DECORATION,
-                    'MARK_SHIFT': MARK,
-                    'MARK_MASK': MARK_MASK,
-            }.items():
-                vv = vv.replace(f'{{{gln}}}', str(pyn), 1)
+            ff = f.replace('{WHICH_PROGRAM}', which)
+            vv = multi_replace(
+                v,
+                WHICH_PROGRAM=which,
+                REVERSE_SHIFT=REVERSE,
+                STRIKE_SHIFT=STRIKETHROUGH,
+                DIM_SHIFT=DIM,
+                DECORATION_SHIFT=DECORATION,
+                MARK_SHIFT=MARK,
+                MARK_MASK=MARK_MASK,
+                DECORATION_MASK=DECORATION_MASK,
+                STRIKE_SPRITE_INDEX=NUM_UNDERLINE_STYLES + 1,
+            )
             if semi_transparent:
                 vv = vv.replace('#define NOT_TRANSPARENT', '#define TRANSPARENT')
                 ff = ff.replace('#define NOT_TRANSPARENT', '#define TRANSPARENT')
@@ -265,8 +355,25 @@ def setup_colors(screen: Screen, opts: Options) -> None:
     )
 
 
+@run_once
+def load_paste_filter() -> Callable[[str], str]:
+    import runpy
+    import traceback
+    try:
+        m = runpy.run_path(os.path.join(config_dir, 'paste-actions.py'))
+        func: Callable[[str], str] = m['filter_paste']
+    except Exception as e:
+        if not isinstance(e, FileNotFoundError):
+            traceback.print_exc()
+            log_error(f'Failed to load paste filter function with error: {e}')
+
+        def func(text: str) -> str:
+            return text
+    return func
+
+
 def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str]:
-    pat = sgr_sanitizer_pat()
+    pat = kitty_ansi_sanitizer_pat()
     ansi, wrap_markers = not as_ansi, not add_wrap_markers
 
     def remove_wrap_markers(line: str) -> str:
@@ -283,6 +390,28 @@ def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str
     return remove_wrap_markers
 
 
+def cmd_output(screen: Screen, which: CommandOutput = CommandOutput.last_run, as_ansi: bool = False, add_wrap_markers: bool = False) -> str:
+    lines: List[str] = []
+    search_in_pager_hist = screen.cmd_output(which, lines.append, as_ansi, add_wrap_markers)
+    if search_in_pager_hist:
+        pht = pagerhist(screen, as_ansi, add_wrap_markers, True)
+        if pht:
+            lines.insert(0, pht)
+    for i in range(min(len(lines), 3)):
+        x = lines[i]
+        if x.startswith('\x1b]133;C'):
+            lines[i] = x.partition('\\')[-1]
+    return ''.join(lines)
+
+
+def process_remote_print(msg: str) -> str:
+    from base64 import standard_b64decode
+
+    from .cli import green
+    text = standard_b64decode(msg).decode('utf-8', 'replace')
+    return text.replace('\x1b', green(r'\e')).replace('\a', green(r'\a')).replace('\0', green(r'\0'))
+
+
 class EdgeWidths:
     left: Optional[float]
     top: Optional[float]
@@ -344,10 +473,14 @@ class Window:
             self.watchers.add(global_watchers())
         else:
             self.watchers = global_watchers().copy()
+        self.last_focused_at = 0.
+        self.started_at = monotonic()
+        self.current_remote_data: List[str] = []
         self.current_mouse_event_button = 0
         self.current_clipboard_read_ask: Optional[bool] = None
         self.prev_osc99_cmd = NotificationCommand()
         self.actions_on_close: List[Callable[['Window'], None]] = []
+        self.actions_on_focus_change: List[Callable[['Window', bool], None]] = []
         self.actions_on_removal: List[Callable[['Window'], None]] = []
         self.current_marker_spec: Optional[Tuple[str, Union[str, Tuple[Tuple[int, str], ...]]]] = None
         self.pty_resized_once = False
@@ -361,6 +494,7 @@ class Window:
         self.id: int = add_window(tab.os_window_id, tab.id, self.title)
         self.margin = EdgeWidths()
         self.padding = EdgeWidths()
+        self.kitten_result: Optional[Dict[str, Any]] = None
         if not self.id:
             raise Exception(f'No tab with id: {tab.id} in OS Window: {tab.os_window_id} was found, or the window counter wrapped')
         self.tab_id = tab.id
@@ -446,14 +580,13 @@ class Window:
         return self.override_title or self.child_title
 
     def __repr__(self) -> str:
-        return 'Window(title={}, id={})'.format(
-                self.title, self.id)
+        return f'Window(title={self.title}, id={self.id})'
 
     def as_dict(self, is_focused: bool = False, is_self: bool = False) -> WindowDict:
         return dict(
             id=self.id,
             is_focused=is_focused,
-            title=self.override_title or self.title,
+            title=self.title,
             pid=self.child.pid,
             cwd=self.child.current_cwd or self.child.cwd,
             cmdline=self.child.cmdline,
@@ -484,9 +617,13 @@ class Window:
     def current_colors(self) -> Dict[str, Optional[int]]:
         return self.screen.color_profile.as_dict()
 
+    @property
+    def at_prompt(self) -> bool:
+        return self.screen.cursor_at_prompt()
+
     @property
     def has_running_program(self) -> bool:
-        return not self.screen.cursor_at_prompt()
+        return not self.at_prompt
 
     def matches(self, field: str, pat: MatchPatternType) -> bool:
         if not pat:
@@ -502,9 +639,9 @@ class Window:
         assert not isinstance(pat, tuple)
 
         if field in ('id', 'window_id'):
-            return True if pat.pattern == str(self.id) else False
+            return pat.pattern == str(self.id)
         if field == 'pid':
-            return True if pat.pattern == str(self.child.pid) else False
+            return pat.pattern == str(self.child.pid)
         if field == 'title':
             return pat.search(self.override_title or self.title) is not None
         if field in 'cwd':
@@ -516,6 +653,33 @@ class Window:
             return False
         return False
 
+    def matches_query(self, field: str, query: str, active_tab: Optional[TabType] = None) -> bool:
+        if field in ('num', 'recent'):
+            if active_tab is not None:
+                try:
+                    q = int(query)
+                except Exception:
+                    return False
+                with suppress(Exception):
+                    if field == 'num':
+                        return active_tab.get_nth_window(q) is self
+                    return active_tab.nth_active_window_id(q) == self.id
+            return False
+        if field == 'state':
+            if query == 'active':
+                return active_tab is not None and self is active_tab.active_window
+            if query == 'focused':
+                return active_tab is not None and self is active_tab.active_window and current_os_window() == self.os_window_id
+            if query == 'needs_attention':
+                return self.needs_attention
+            if query == 'parent_active':
+                return active_tab is not None and self.tabref() is active_tab
+            if query == 'parent_focused':
+                return active_tab is not None and self.tabref() is active_tab and current_os_window() == self.os_window_id
+            return False
+        pat = compile_match_query(query, field != 'env')
+        return self.matches(field, pat)
+
     def set_visible_in_layout(self, val: bool) -> None:
         val = bool(val)
         if val is not self.is_visible_in_layout:
@@ -587,6 +751,8 @@ class Window:
 
     def write_to_child(self, data: Union[str, bytes]) -> None:
         if data:
+            if isinstance(data, str):
+                data = data.encode('utf-8')
             if get_boss().child_monitor.needs_write(self.id, data) is not True:
                 log_error(f'Failed to write to child {self.id} as it does not exist')
 
@@ -649,32 +815,38 @@ class Window:
                     url = urlunparse(purl._replace(netloc=''))
             if opts.allow_hyperlinks & 0b10:
                 from kittens.tui.operations import styled
-                get_boss()._run_kitten('ask', ['--type=choices', '--message', _(
-                    'What would you like to do with this URL:\n') +
-                    styled(unquote(url), fg='yellow'),
-                    '--choice=o:Open', '--choice=c:Copy to clipboard', '--choice=n;red:Nothing'
-                    ],
+                get_boss().choose(
+                    'What would you like to do with this URL:\n' + styled(unquote(url), fg='yellow'),
+                    partial(self.hyperlink_open_confirmed, url, cwd),
+                    'o:Open', 'c:Copy to clipboard', 'n;red:Nothing', default='o',
                     window=self,
-                    custom_callback=partial(self.hyperlink_open_confirmed, url, cwd),
-                    default_data={'response': ''}
                 )
                 return
         get_boss().open_url(url, cwd=cwd)
 
-    def hyperlink_open_confirmed(self, url: str, cwd: Optional[str], data: Dict[str, Any], *a: Any) -> None:
-        q = data.get('response', '')
+    def hyperlink_open_confirmed(self, url: str, cwd: Optional[str], q: str) -> None:
         if q == 'o':
             get_boss().open_url(url, cwd=cwd)
         elif q == 'c':
             set_clipboard_string(url)
 
     def handle_remote_file(self, netloc: str, remote_path: str) -> None:
+        from .utils import SSHConnectionData
         from kittens.ssh.main import get_connection_data
-        args = self.child.foreground_cmdline
-        conn_data = get_connection_data(args, self.child.foreground_cwd or self.child.current_cwd or '')
+        from kittens.remote_file.main import is_ssh_kitten_sentinel
+        args = self.ssh_kitten_cmdline()
+        conn_data: Union[None, List[str], SSHConnectionData] = None
+        if args:
+            ssh_cmdline = sorted(self.child.foreground_processes, key=lambda p: p['pid'])[-1]['cmdline'] or ['']
+            if 'ControlPath=' in ' '.join(ssh_cmdline):
+                idx = ssh_cmdline.index('--')
+                conn_data = [is_ssh_kitten_sentinel] + list(ssh_cmdline[:idx + 2])
         if conn_data is None:
-            get_boss().show_error('Could not handle remote file', 'No SSH connection data found in: {args}')
-            return
+            args = self.child.foreground_cmdline
+            conn_data = get_connection_data(args, self.child.foreground_cwd or self.child.current_cwd or '')
+            if conn_data is None:
+                get_boss().show_error('Could not handle remote file', f'No SSH connection data found in: {args}')
+                return
         get_boss().run_kitten(
             'remote_file', '--hostname', netloc.partition(':')[0], '--path', remote_path,
             '--ssh-connection-data', json.dumps(conn_data)
@@ -684,18 +856,28 @@ class Window:
         if self.destroyed:
             return
         call_watchers(weakref.ref(self), 'on_focus_change', {'focused': focused})
+        for c in self.actions_on_focus_change:
+            try:
+                c(self, focused)
+            except Exception:
+                import traceback
+                traceback.print_exc()
         self.screen.focus_changed(focused)
         if focused:
-            update_ime_position_for_window(self.id)
+            self.last_focused_at = monotonic()
+            update_ime_position_for_window(self.id, False, 1)
             changed = self.needs_attention
             self.needs_attention = False
             if changed:
                 tab = self.tabref()
                 if tab is not None:
                     tab.relayout_borders()
+        elif self.os_window_id == current_os_window():
+            # Cancel IME composition after loses focus
+            update_ime_position_for_window(self.id, False, -1)
 
-    def title_changed(self, new_title: Optional[str]) -> None:
-        self.child_title = sanitize_title(new_title or self.default_title)
+    def title_changed(self, new_title: Optional[str], is_base64: bool = False) -> None:
+        self.child_title = process_title_from_child(new_title or self.default_title, is_base64)
         if self.override_title is None:
             self.title_updated()
 
@@ -719,7 +901,7 @@ class Window:
         if cb and cb != ['none']:
             import shlex
             import subprocess
-            env = self.child.final_env
+            env = self.child.foreground_environ
             env['KITTY_CHILD_CMDLINE'] = ' '.join(map(shlex.quote, self.child.cmdline))
             subprocess.Popen(cb, env=env, cwd=self.child.foreground_cwd)
         if not self.is_active:
@@ -765,6 +947,10 @@ class Window:
         if default_bg_changed:
             get_boss().default_bg_changed_for(self.id)
 
+    def color_profile_popped(self, bg_changed: bool) -> None:
+        if bg_changed:
+            get_boss().default_bg_changed_for(self.id)
+
     def report_color(self, code: str, r: int, g: int, b: int) -> None:
         r |= r << 8
         g |= g << 8
@@ -825,12 +1011,93 @@ class Window:
     def handle_remote_cmd(self, cmd: str) -> None:
         get_boss().handle_remote_cmd(cmd, self)
 
-    def handle_remote_print(self, msg: bytes) -> None:
+    def handle_remote_echo(self, msg: str) -> None:
         from base64 import standard_b64decode
+        data = standard_b64decode(msg)
+        self.write_to_child(data)
 
-        from .cli import green
-        text = standard_b64decode(msg).decode('utf-8')
-        text = text.replace('\x1b', green(r'\e')).replace('\a', green(r'\a')).replace('\0', green(r'\0'))
+    def handle_remote_ssh(self, msg: str) -> None:
+        from kittens.ssh.main import get_ssh_data
+        for line in get_ssh_data(msg, f'{os.getpid()}-{self.id}'):
+            self.write_to_child(line)
+
+    def handle_kitten_result(self, msg: str) -> None:
+        import base64
+        self.kitten_result = json.loads(base64.b85decode(msg))
+
+    def handle_overlay_ready(self, msg: str) -> None:
+        boss = get_boss()
+        tab = boss.tab_for_window(self)
+        if tab is not None:
+            tab.move_window_to_top_of_group(self)
+
+    def append_remote_data(self, msg: str) -> str:
+        if not msg:
+            cdata = ''.join(self.current_remote_data)
+            self.current_remote_data = []
+            return cdata
+        num, rest = msg.split(':', 1)
+        max_size = get_options().clipboard_max_size * 1024 * 1024
+        if num == '0' or sum(map(len, self.current_remote_data)) > max_size:
+            self.current_remote_data = []
+        self.current_remote_data.append(rest)
+        return ''
+
+    def handle_remote_edit(self, msg: str) -> None:
+        cdata = self.append_remote_data(msg)
+        if cdata:
+            from .launch import remote_edit
+            remote_edit(cdata, self)
+
+    def handle_remote_clone(self, msg: str) -> None:
+        cdata = self.append_remote_data(msg)
+        if cdata:
+            ac = get_options().allow_cloning
+            if ac == 'ask':
+                get_boss().confirm(_(
+                    'A program running in this window wants to clone it into another window.'
+                    ' Allow it do so, once?'),
+                    partial(self.handle_remote_clone_confirmation, cdata), window=self,
+                )
+            elif ac in ('yes', 'y', 'true'):
+                self.handle_remote_clone_confirmation(cdata, True)
+
+    def handle_remote_clone_confirmation(self, cdata: str, confirmed: bool) -> None:
+        if confirmed:
+            from .launch import clone_and_launch
+            clone_and_launch(cdata, self)
+
+    def handle_remote_askpass(self, msg: str) -> None:
+        from .shm import SharedMemory
+        with SharedMemory(name=msg, readonly=True) as shm:
+            shm.seek(1)
+            data = json.loads(shm.read_data_with_size())
+
+        def callback(ans: Any) -> None:
+            data = json.dumps(ans)
+            with SharedMemory(name=msg) as shm:
+                shm.seek(1)
+                shm.write_data_with_size(data)
+                shm.flush()
+                shm.seek(0)
+                shm.write(b'\x01')
+
+        message: str = data['message']
+        if data['type'] == 'confirm':
+            get_boss().confirm(
+                message, callback, window=self, confirm_on_cancel=bool(data.get('confirm_on_cancel')),
+                confirm_on_accept=bool(data.get('confirm_on_accept', True)))
+        elif data['type'] == 'choose':
+            get_boss().choose(
+                message, callback, *data['choices'], window=self, default=data.get('default', ''))
+        elif data['type'] == 'get_line':
+            get_boss().get_line(
+                message, callback, window=self, is_password=bool(data.get('is_password')), prompt=data.get('prompt', '> '))
+        else:
+            log_error(f'Ignoring ask request with unknown type: {data["type"]}')
+
+    def handle_remote_print(self, msg: str) -> None:
+        text = process_remote_print(msg)
         print(text, end='', file=sys.stderr)
         sys.stderr.flush()
 
@@ -954,6 +1221,12 @@ class Window:
                 if click_mouse_url(self.os_window_id, self.tab_id, self.id):
                     break
             if a == 'prompt':
+                # Do not send move cursor events too soon after the window is
+                # focused, this is because there are people that click on
+                # windows and start typing immediately and the cursor event
+                # can interfere with that. See https://github.com/kovidgoyal/kitty/issues/4128
+                if monotonic() - self.last_focused_at < 1.5 * get_click_interval():
+                    return
                 if move_cursor_to_mouse_if_in_prompt(self.os_window_id, self.tab_id, self.id):
                     self.screen.ignore_bells_for(1)
                     break
@@ -978,13 +1251,13 @@ class Window:
     def paste_selection(self) -> None:
         txt = get_boss().current_primary_selection()
         if txt:
-            self.paste(txt)
+            self.paste_with_actions(txt)
 
     @ac('mouse', 'Paste the current primary selection or the clipboard if no selection is present')
     def paste_selection_or_clipboard(self) -> None:
         txt = get_boss().current_primary_selection_or_clipboard()
         if txt:
-            self.paste(txt)
+            self.paste_with_actions(txt)
 
     @ac('mouse', '''
         Select clicked command output
@@ -1004,12 +1277,10 @@ class Window:
             self.show_cmd_output(CommandOutput.last_visited, 'Clicked command output')
     # }}}
 
-    def text_for_selection(self) -> str:
-        lines = self.screen.text_for_selection()
+    def text_for_selection(self, as_ansi: bool = False) -> str:
         sts = get_options().strip_trailing_spaces
-        if sts == 'always' or (
-                sts == 'smart' and not self.screen.is_rectangle_select()):
-            return ''.join((ln.rstrip() or '\n') for ln in lines)
+        strip_trailing_spaces = sts == 'always' or (sts == 'smart' and not self.screen.is_rectangle_select())
+        lines = self.screen.text_for_selection(as_ansi, strip_trailing_spaces)
         return ''.join(lines)
 
     def call_watchers(self, which: Iterable[Watcher], data: Dict[str, Any]) -> None:
@@ -1025,6 +1296,9 @@ class Window:
         self.call_watchers(self.watchers.on_close, {})
         self.destroyed = True
         if hasattr(self, 'screen'):
+            if self.is_active and self.os_window_id == current_os_window():
+                # Cancel IME composition when window is destroyed
+                update_ime_position_for_window(self.id, False, -1)
             # Remove cycles so that screen is de-allocated immediately
             self.screen.reset_callbacks()
             del self.screen
@@ -1040,13 +1314,30 @@ class Window:
         return as_text(self.screen, as_ansi, add_history, add_wrap_markers, alternate_screen, add_cursor)
 
     def cmd_output(self, which: CommandOutput = CommandOutput.last_run, as_ansi: bool = False, add_wrap_markers: bool = False) -> str:
-        lines: List[str] = []
-        self.screen.cmd_output(which, lines.append, as_ansi, add_wrap_markers)
-        return ''.join(lines)
+        return cmd_output(self.screen, which, as_ansi, add_wrap_markers)
+
+    def get_cwd_of_child(self, oldest: bool = False) -> Optional[str]:
+        return self.child.get_foreground_cwd(oldest) or self.child.current_cwd
 
     @property
     def cwd_of_child(self) -> Optional[str]:
-        return self.child.foreground_cwd or self.child.current_cwd
+        return self.get_cwd_of_child()
+
+    @property
+    def child_is_remote(self) -> bool:
+        for p in self.child.foreground_processes:
+            q = list(p['cmdline'] or ())
+            if q and q[0].lower() == 'ssh':
+                return True
+        return False
+
+    def ssh_kitten_cmdline(self) -> List[str]:
+        from kittens.ssh.utils import is_kitten_cmdline
+        for p in self.child.foreground_processes:
+            q = list(p['cmdline'] or ())
+            if is_kitten_cmdline(q):
+                return q
+        return []
 
     def pipe_data(self, text: str, has_wrap_markers: bool = False) -> PipeData:
         text = text or ''
@@ -1068,6 +1359,68 @@ class Window:
         path = resolve_custom_file(path) if path else ''
         set_window_logo(self.os_window_id, self.tab_id, self.id, path, position or '', alpha)
 
+    def paste_with_actions(self, text: str) -> None:
+        if self.destroyed or not text:
+            return
+        opts = get_options()
+        if 'filter' in opts.paste_actions:
+            text = load_paste_filter()(text)
+            if not text:
+                return
+        if 'quote-urls-at-prompt' in opts.paste_actions and self.at_prompt:
+            prefixes = '|'.join(opts.url_prefixes)
+            m = re.match(f'({prefixes}):(.+)', text)
+            if m is not None:
+                scheme, rest = m.group(1), m.group(2)
+                if rest.startswith('//') or scheme in ('mailto', 'irc'):
+                    import shlex
+                    text = shlex.quote(text)
+        btext = text.encode('utf-8')
+        if 'confirm' in opts.paste_actions:
+            msg = ''
+            limit = 16 * 1024
+            if not self.screen.in_bracketed_paste_mode:
+                msg = _('Pasting text into shells that do not support bracketed paste can be dangerous.')
+            elif len(btext) > limit:
+                msg = _('Pasting very large amounts of text ({} bytes) can be slow.').format(len(btext))
+            if msg:
+                get_boss().confirm(msg + _(' Are you sure?'), partial(self.handle_paste_confirmation, btext), window=self)
+                return
+        self.paste_text(btext)
+
+    def handle_paste_confirmation(self, btext: bytes, confirmed: bool) -> None:
+        if confirmed:
+            self.paste_text(btext)
+
+    def paste_bytes(self, text: Union[str, bytes]) -> None:
+        # paste raw bytes without any processing
+        if isinstance(text, str):
+            text = text.encode('utf-8')
+        self.screen.paste_bytes(text)
+
+    def paste_text(self, text: Union[str, bytes]) -> None:
+        if text and not self.destroyed:
+            if isinstance(text, str):
+                text = text.encode('utf-8')
+            if self.screen.in_bracketed_paste_mode:
+                while True:
+                    new_text = text.replace(b'\033[201~', b'').replace(b'\x9b201~', b'')
+                    if len(text) == len(new_text):
+                        break
+                    text = new_text
+            else:
+                # Workaround for broken editors like nano that cannot handle
+                # newlines in pasted text see https://github.com/kovidgoyal/kitty/issues/994
+                text = text.replace(b'\r\n', b'\n').replace(b'\n', b'\r')
+            self.screen.paste(text)
+
+    def clear_screen(self, reset: bool = False, scrollback: bool = False) -> None:
+        self.screen.cursor.x = self.screen.cursor.y = 0
+        if reset:
+            self.screen.reset()
+        else:
+            self.screen.erase_in_display(3 if scrollback else 2, False)
+
     # actions {{{
 
     @ac('cp', 'Show scrollback in a pager like less')
@@ -1107,28 +1460,17 @@ class Window:
     def show_last_visited_command_output(self) -> None:
         self.show_cmd_output(CommandOutput.last_visited, 'Last visited command output')
 
-    def paste_bytes(self, text: Union[str, bytes]) -> None:
-        # paste raw bytes without any processing
-        if isinstance(text, str):
-            text = text.encode('utf-8')
-        self.screen.paste_bytes(text)
+    @ac('cp', '''
+        Show the last non-empty output from a shell command in a pager like less
+
+        Requires :ref:`shell_integration` to work
+        ''')
+    def show_last_non_empty_command_output(self) -> None:
+        self.show_cmd_output(CommandOutput.last_non_empty, 'Last non-empty command output')
 
     @ac('cp', 'Paste the specified text into the current window')
-    def paste(self, text: Union[str, bytes]) -> None:
-        if text and not self.destroyed:
-            if isinstance(text, str):
-                text = text.encode('utf-8')
-            if self.screen.in_bracketed_paste_mode:
-                while True:
-                    new_text = text.replace(b'\033[201~', b'').replace(b'\x9b201~', b'')
-                    if len(text) == len(new_text):
-                        break
-                    text = new_text
-            else:
-                # Workaround for broken editors like nano that cannot handle
-                # newlines in pasted text see https://github.com/kovidgoyal/kitty/issues/994
-                text = text.replace(b'\r\n', b'\n').replace(b'\n', b'\r')
-            self.screen.paste(text)
+    def paste(self, text: str) -> None:
+        self.paste_with_actions(text)
 
     @ac('cp', 'Copy the selected text from the active window to the clipboard')
     def copy_to_clipboard(self) -> None:
@@ -1136,6 +1478,12 @@ class Window:
         if text:
             set_clipboard_string(text)
 
+    @ac('cp', 'Copy the selected text from the active window to the clipboard with ANSI formatting codes')
+    def copy_ansi_to_clipboard(self) -> None:
+        text = self.text_for_selection(as_ansi=True)
+        if text:
+            set_clipboard_string(text)
+
     def encoded_key(self, key_event: KeyEvent) -> bytes:
         return encode_key_for_tty(
             key=key_event.key, shifted_key=key_event.shifted_key, alternate_key=key_event.alternate_key,
@@ -1144,15 +1492,16 @@ class Window:
             cursor_key_mode=self.screen.cursor_key_mode,
         ).encode('ascii')
 
-    @ac('cp', 'Copy the selected text from the active window to the clipboard, if no selection, send Ctrl-C')
+    @ac('cp', 'Copy the selected text from the active window to the clipboard, if no selection, send SIGINT (aka :kbd:`ctrl+c`)')
     def copy_or_interrupt(self) -> None:
         text = self.text_for_selection()
         if text:
             set_clipboard_string(text)
         else:
+            self.scroll_end()
             self.write_to_child(self.encoded_key(KeyEvent(key=ord('c'), mods=GLFW_MOD_CONTROL)))
 
-    @ac('cp', 'Copy the selected text from the active window to the clipboard and clear selection, if no selection, send Ctrl-C')
+    @ac('cp', 'Copy the selected text from the active window to the clipboard and clear selection, if no selection, send SIGINT (aka :kbd:`ctrl+c`)')
     def copy_and_clear_or_interrupt(self) -> None:
         self.copy_or_interrupt()
         self.screen.clear_selection()
@@ -1167,6 +1516,10 @@ class Window:
             else:
                 open_url(text, cwd=cwd)
 
+    @ac('cp', 'Clear the current selection')
+    def clear_selection(self) -> None:
+        self.screen.clear_selection()
+
     @ac('sc', 'Scroll up by one line')
     def scroll_line_up(self) -> None:
         if self.screen.is_main_linebuf():
@@ -1213,6 +1566,19 @@ class Window:
         if self.screen.is_main_linebuf():
             self.screen.scroll_to_prompt(num_of_prompts)
 
+    @ac('sc', 'Scroll prompt to the top of the screen, filling screen with empty lines')
+    def scroll_prompt_to_top(self, clear_scrollback: bool = False) -> None:
+        if self.screen.is_main_linebuf():
+            self.screen.scroll_until_cursor_prompt()
+            if clear_scrollback:
+                self.screen.clear_scrollback()
+            elif self.screen.scrolled_by > 0:
+                self.screen.scroll(SCROLL_FULL, False)
+
+    @ac('sc', 'Scroll prompt to the bottom of the screen, filling in extra lines from the scrollback buffer')
+    def scroll_prompt_to_bottom(self) -> None:
+        self.screen.scroll_prompt_to_bottom()
+
     @ac('mk', 'Toggle the current marker on/off')
     def toggle_marker(self, ftype: str, spec: Union[str, Tuple[Tuple[int, str], ...]], flags: int) -> None:
         from .marks import marker_from_spec
